#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : argument.py
# Author : Jiayuan Mao
# Email  : maojiayuan@gmail.com
# Date   : 03/14/2017
#
# This file is part of Jacinle.
# Distributed under terms of the MIT license.

import collections

__all__ = [
    "get_2dshape",
    "get_3dshape",
    "get_4dshape",
    "astuple",
    "asshape",
    "canonize_args_list",
    "UniqueValueGetter",
]


def get_2dshape(x, default=None, type=int):
    """Convert a value or a tuple to a tuple of length 2.
    
    Args:
        x: a value of type `type`, or a tuple of length 2.
        default: default value.
        type: expected type of the element.

    Returns: a tuple of length 2.
    """
    if x is None:
        return default
    if isinstance(x, collections.abc.Sequence):
        x = tuple(x)
        if len(x) == 1:
            return x[0], x[0]
        else:
            assert len(x) == 2, "2dshape must be of length 1 or 2"
            return x
    else:
        x = type(x)
        return x, x


def get_3dshape(x, default=None, type=int):
    if x is None:
        return default
    if isinstance(x, collections.abc.Sequence):
        x = tuple(x)
        if len(x) == 1:
            return x[0], x[0], x[0]
        else:
            assert len(x) == 3, "3dshape must be of length 1 or 3"
            return x
    else:
        x = type(x)
        return x, x, x


def get_4dshape(x, default=None, type=int):
    if x is None:
        return default
    if isinstance(x, collections.abc.Sequence):
        x = tuple(x)
        if len(x) == 1:
            return 1, x[0], x[0], 1
        elif len(x) == 2:
            return 1, x[0], x[1], 1
        else:
            assert len(x) == 4, "4dshape must be of length 1, 2, or 4"
            return x
    else:
        x = type(x)
        return 1, x, x, 1


def astuple(arr_like):
    """Convert a sequence or a single value to a tuple. This method differ from the system method `tuple` in that
    a single value (incl. int, string, bytes) will be converted to a tuple of length 1.

    Args:
        arr_like: a sequence or a single value.

    Returns: a tuple.
    """
    if type(arr_like) is tuple:
        return arr_like
    elif isinstance(arr_like, collections.abc.Sequence) and not isinstance(
        arr_like, (str, bytes)
    ):
        return tuple(arr_like)
    else:
        return tuple((arr_like,))


def asshape(arr_like):
    if type(arr_like) is tuple:
        return arr_like
    elif type(arr_like) is int:
        if arr_like == 0:
            return tuple()
        else:
            return tuple((arr_like,))
    elif arr_like is None:
        return (None,)
    else:
        return tuple(arr_like)


def canonize_args_list(args, *, allow_empty=False, cvt=None):
    """Convert the argument list to a tuple of values. This is useful to make unified interface for shape-related
    operations. E.g.,

    >>> np.zeros(6).reshape(2, 3)
    >>> np.zeros(6).reshape((2, 3))

    Args:
        args:
        allow_empty:
        cvt:

    Returns:

    """
    if not allow_empty and not args:
        raise TypeError("at least one argument must be provided")

    if len(args) == 1 and isinstance(args[0], (list, tuple)):
        args = args[0]
    if cvt is not None:
        args = tuple(map(cvt, args))
    return args


class UniqueValueGetter(object):
    def __init__(self, msg="unique value check failed", default=None):
        self._msg = msg
        self._val = None
        self._default = default

    def set(self, v):
        assert (
            self._val is None or self._val == v
        ), self._msg + ": expect={} got={}".format(self._val, v)
        self._val = v

    def get(self):
        return self._val or self._default



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : naming.py
# Author : Jiayuan Mao
# Email  : maojiayuan@gmail.com
# Date   : 01/18/2017
#
# This file is part of Jacinle.
# Distributed under terms of the MIT license.

__all__ = ["class_name", "func_name", "method_name", "class_name_of_method"]


def class_name(instance_or_class):
    if isinstance(instance_or_class, type):
        return func_name(instance_or_class)
    return func_name(instance_or_class.__class__)


def func_name(func):
    return func.__module__ + "." + func.__qualname__


def method_name(method):
    assert "." in method.__qualname__, '"{}" is not a method.'.format(repr(method))
    return func_name(method)


def class_name_of_method(method):
    name = method_name(method)
    return name[: name.rfind(".")]



import torch
from jamtorch.trainer import GeneticTrainer
from omegaconf import DictConfig
import torch.distributed as dist
from jammy.utils.hyd import hyd_instantiate
from torch.nn.parallel import DistributedDataParallel as DDP
from jamtorch.utils.meta import is_master
import jammy.utils.hyd as hyd
import jamtorch.trainer.progress_fn as progress_fn
from jamtorch.io import attr_dict
import os.path as osp
import tempfile

from jamtorch.logging import get_logger

logger = get_logger()


class DDPTrainer(GeneticTrainer):
    def __init__(self, cfg, loss_fn):
        super().__init__(cfg, loss_fn)
        self.train_sampler, self.val_sampler = None, None
        self.is_master = is_master()
        self.rank = 0 if self.is_master else dist.get_rank()
        self.setup_ddp()

    def set_model_optim(self, model, optimizer):
        # assert isinstance(optimizer_cfg, DictConfig)
        if next(model.parameters()).device == torch.device("cpu"):
            model = model.to(self.device)
        if self._cfg.dist.syncBN:
            model = torch.nn.SyncBatchNorm.convert_sync_batchnorm(model)
        model = DDP(model, device_ids=[self.rank])
        # instantiate optmiizer after DDP model
        # optimizer = hyd_instantiate(optimizer_cfg, model.parameters())
        super().set_model_optim(model, optimizer)

    def init_model(self):
        if self._cfg.resume:
            return
        checkpoint_path = osp.join(tempfile.gettempdir(), "initial_weights.pt")
        if self.is_master:
            torch.save(self.model.state_dict(), checkpoint_path)

        dist.barrier()
        self.model.load_state_dict(
            torch.load(checkpoint_path, map_location=self.device)
        )

    def set_sampler(self, train_sampler, val_sampler):
        self.train_sampler, self.val_sampler = train_sampler, val_sampler

    def setup_ddp(self):
        self.register_event(
            "epoch:before",
            lambda trainer: trainer.train_sampler.set_epoch(self.epoch_cnt),
            False,
        )
        if self.is_master:
            progress_fn.simple_train_bar(self)
            progress_fn.simple_val_bar(self)

    def _impl_load_ckpt(self, state):
        """
        only load model, bypass optimizer
        """
        # The creatation of optimizer needs wait after model
        msg_model = self.model.module.load_state_dict(state["model"])
        logger.critical(f"load model ckpt {msg_model}")
        self._impl_load_amp_scaler(state)

        dist.barrier()

    def save_ckpt(self, val_loss=float("inf")):
        if self.is_master:
            return super().save_ckpt(val_loss)
        return

    def _impl_save_ckpt(self):
        super_ckpt = attr_dict(self, ["model", "amp_scaler"])
        return super_ckpt



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : memcached.py
# Author : Jiayuan Mao
# Email  : maojiayuan@gmail.com
# Date   : 10/23/2018
#
# This file is part of Jacinle.
# Distributed under terms of the MIT license.

import collections

from jammy.storage.kv.memcached import MemcachedKVStore

from .session import SessionManagerBase

__all__ = ["MemcachedSessionManager", "MemcachedSessionIdentifier"]


class MemcachedSessionIdentifier(
    collections.namedtuple("_MemcachedSessionIdentifier", ["session_id", "hmac_key"])
):
    pass


class MemcachedSessionManager(SessionManagerBase):
    def __init__(
        self,
        secret,
        memcache_host,
        memcache_port,
        timeout,
        cookie_prefix="jam_ses_",
        memcached_prefix="jam_sess_",
    ):  # pylint: disable=too-many-arguments
        super().__init__(secret)

        self.memcache = MemcachedKVStore(memcache_host, memcache_port)
        self.session_timeout = timeout
        self.cookie_prefix = cookie_prefix
        self.memcached_prefix = memcached_prefix

        assert self.memcache.available

    def get(self, request_handler):
        session_id = hmac_key = None
        if request_handler is not None:
            session_id = request_handler.get_secure_cookie(
                self.cookie_prefix + "session_id"
            ).decode("utf8")
            hmac_key = request_handler.get_secure_cookie(
                self.cookie_prefix + "verification"
            ).decode("utf8")

        if session_id is None:
            session_exists = False
            session_id = self._generate_id()
            hmac_key = self._generate_hmac(session_id)
        else:
            session_exists = True

        data = {}
        if session_exists and hmac_key == self._generate_hmac(session_id):
            data = self.memcache.get(
                self.memcached_prefix + session_id,
                data,
                refresh=True,
                refresh_timeout=self.session_timeout,
            )
        return MemcachedSessionIdentifier(session_id, hmac_key), data

    def set(
        self, request_handler, identifier, data
    ):  # pylint: disable=arguments-differ
        request_handler.set_secure_cookie(
            self.cookie_prefix + "session_id", identifier.session_id.encode("utf8")
        )
        request_handler.set_secure_cookie(
            self.cookie_prefix + "verification", identifier.hmac_key.encode("utf8")
        )
        self.memcache.put(
            self.memcached_prefix + identifier.session_id,
            data,
            replace=True,
            timeout=self.session_timeout,
        )



#!/usr/bin/env python3
# pylint: disable=global-variable-not-assigned, global-statement
import os
import os.path as osp
import json
import time
import hydra
from omegaconf import DictConfig, OmegaConf
from jammy.cli.keyboard import yes_or_no

from jammy.logging import get_logger
from jammy.utils.process import run_simple_command

logger = get_logger()

job_cmd = []
job_cfg = {
    'aceId': 257,
    'aceInstance': 'dgx1v.16g.1.norm',
    'aceName': 'nv-us-west-2',
    'name': 'ml-model.qsh.16g.1_debug',
    'publishedContainerPorts': [9999, 8888],
    'dockerImageName': 'nvidian/lpr-imagine/imaginaire_qsh:1.1',
    'command': 'cd /mnt/qsh_ws;tmux new-session -d -s dv \"nvitop\"; sleep 7d',
    'minAvailability': 1,
    'replicaCount': 1,
    'arrayType': 'PYTORCH',
    'runPolicy': {'totalRuntimeSeconds': 604800, 'preemptClass': 'RUNONCE'},
    'resultContainerMountPoint': '/result',
    'workspaceMounts': [
    ],
}

def parse_all(cfg):
    if cfg.img is not None:
        job_cfg['dockerImageName'] = cfg.img
    if cfg.name is not None:
        job_cfg['name'] = "ml-model.qsh_debug."+cfg.name
    else:
        job_cfg['name'] = f"ml-model.qsh_debug.{cfg.device}"
    parse_device(cfg)
    parse_cmd(cfg)
    parse_ws_qsh(cfg)
    parse_extra_ws(cfg)

def parse_device(cfg):
    if cfg.device is not None:
        cfg.device = str(cfg.device)
        if "x" in cfg.device:
            if cfg.device.count("x") == 1:
                num_node = 1
                gpu_mem = cfg.device.split("x")[0]
                gpu_per_node = int(cfg.device.split("x")[1])
            elif cfg.device.count("x") == 2:
                num_node = int(cfg.device.split("x")[0])
                gpu_mem = cfg.device.split("x")[1]
                gpu_per_node = int(cfg.device.split("x")[2])
            elif cfg.device.count("x") > 2:
                raise ValueError(f"device format error {cfg.device}")
        else:
            num_node = 1
            gpu_mem = cfg.device
            gpu_per_node = 1
        job_cfg["aceInstance"] = {
            "cpu": "cpu.x86.tiny",
            "16": f"dgx1v.16g.{gpu_per_node}.norm",
            "32": f"dgx1v.32g.{gpu_per_node}.norm",
            "32b": f"dgx1v.32g.{gpu_per_node}.norm.beta",
            "40": f"dgxa100.40g.{gpu_per_node}.norm",
            "40b": f"dgxa100.40g.{gpu_per_node}.norm",
            "48": f"ovxa40.48g.{gpu_per_node}",
            "80": f"dgxa100.80g.{gpu_per_node}.norm",
            "80b": f"dgxa100.80g.{gpu_per_node}.norm",
        }[str(gpu_mem)]

    else:
        num_node = 1
        gpu_mem = "cpu"
        gpu_per_node = 1

        job_cfg["aceInstance"] = "cpu.x86.tiny"
    job_cfg["minAvailability"] = job_cfg["replicaCount"] = num_node

    job_ace = {
        "cpu": "nv-us-west-2",
        "16": "nv-us-west-2",
        "32": "nv-us-west-2",
        "32b": "nv-us-west-2",
        "40": "nv-us-east-1",
        "40b": "nv-us-west-2",
        "48": "ov-us-west-2",
        "80": "nv-us-east-1",
        "80b": "nv-us-west-3",
    }[str(gpu_mem)]

    job_team = "deep-imagination" if "dgxa100" in job_cfg["aceInstance"] else "lpr-imagine"
    job_cmd.append(
        f"--team {job_team} --ace {job_ace} --label qsh_debug --priority HIGH --order 1 --preempt RUNONCE"
    )

    cfg.job_ace = job_ace

def parse_cmd(cfg):
    cmds = []
    if cfg.ws_qsh:
        cmds.append("cd /mnt/qsh_ws")
    if cfg.cmd is not None:
        cmd = cfg.cmd
        if type(cmd) is str:
            cmd = [cmd]
        cmds.extend(cmd)
    if cfg.debug:
        cmds.append("tmux new-session -d -s dv \"nvitop\"; sleep 7d")
    job_cfg['command'] = ";".join(cmds)


def parse_ws_qsh(cfg):
    if cfg.key_qsh:
        keys_ws = {
            'nv-us-east-1': 'qsh_keys1',
            'nv-us-west-2': 'qsh_keys',
            'nv-us-west-3': 'qsh_keys3',
        }[cfg.job_ace]
        job_cfg['workspaceMounts'].append({
            "containerMountPoint": "/mnt/qsh_keys",
            "id": keys_ws,
            "mountMode": "RW"
        })
    if cfg.ws_qsh:
        ws = {
            'nv-us-east-1': 'qsh_ws_e1',
            'nv-us-west-2': 'qsh_ws',
            'nv-us-west-3': 'qsh_ws_w3',
        }[cfg.job_ace]
        job_cfg['workspaceMounts'].append({
            "containerMountPoint": "/mnt/qsh_ws",
            "id": ws,
            "mountMode": "RW"
        })

def parse_extra_ws(cfg):
    if cfg.extra_ws is not None:
        for ws, target_loc in cfg.extra_ws.items():
            job_cfg['workspaceMounts'].append({
                "containerMountPoint": target_loc,
                "id": ws,
                "mountMode": "RW"
            })

def read_local_cfg(cfg):
    cfg_path = hydra.utils.to_absolute_path(cfg.file)
    if not osp.exists(cfg_path):
        return cfg
    logger.info("readding config from: " + cfg_path)
    local_cfg = OmegaConf.load(cfg_path)
    cli_cfg = OmegaConf.from_cli()
    cfg = OmegaConf.merge(cfg, local_cfg, cli_cfg)
    return cfg

@hydra.main(config_path="conf", config_name="ngc", version_base="1.1.0")
def my_app(cfg: DictConfig) -> None:
    OmegaConf.set_struct(cfg, False)
    cfg = read_local_cfg(cfg)
    parse_all(cfg)

    job_cfg, job_cmd
    job_name = job_cfg["name"]
    json_fname = f"/tmp/{job_name}.json"
    os.makedirs(os.path.dirname(json_fname), exist_ok=True)
    with open(json_fname, "w") as file:
        json.dump(job_cfg, file, indent=4)

    system_cmd = f"ngc batch run -f {json_fname} " + " ".join(job_cmd)
    logger.info(system_cmd)
    if yes_or_no("submit job? (y/n) ", default="yes"):
        stdout, stderr = run_simple_command(system_cmd)
        if not stderr and cfg.exec:
            id_info = stdout.split('\n')[2]
            job_id = id_info.split(' ')[-1]
            logger.warning(f"ngc batch exec {job_id}")
            while True:
                rtn_value = os.system(f"ngc batch exec {job_id}")
                if rtn_value != 0:
                    logger.warning(f"ERROR VALUE {rtn_value}: ngc batch exec {job_id} failed, retrying...")
                    time.sleep(5)
                else:
                    break
        if stderr:
            print(stderr)


if __name__ == "__main__":
    my_app()  # pylint: disable=no-value-for-parameter



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : monitor.py
# Author : Jiayuan Mao
# Email  : maojiayuan@gmail.com
# Date   : 04/12/2018
#
# This file is part of Jacinle.
# Distributed under terms of the MIT license.

import torch.nn.functional as F
from jamtorch.utils.meta import as_float
from jamtorch.utils.grad import no_grad_func

__all__ = [
    "group_prefix",
    "binary_classification_accuracy",
    "classification_accuracy",
    "regression_accuracy",
    "monitor_param_saturation",
    "monitor_param_rms",
    "monitor_param_gradrms",
    "monitor_param_gradrms_ratio",
]


def group_prefix(mode, monitor):
    return {f"{mode}/{k}": v for k, v in monitor.items()}


@no_grad_func
def binary_classification_accuracy(pred, label, name="", saturation=True):
    if name != "":
        name = "/" + name
    prefix = "accuracy" + name
    pred = pred.view(-1)  # Binary accuracy
    label = label.view(-1)
    acc = label.float().eq((pred > 0.5).float())
    if saturation:
        sat = 1 - (pred - (pred > 0.5).float()).abs()
        return {
            prefix: as_float(acc.float().mean()),
            prefix + "/saturation/mean": as_float(sat.mean()),
            prefix + "/saturation/min": as_float(sat.min()),
        }
    return {prefix: as_float(acc.float().mean())}


@no_grad_func
def classification_accuracy(pred, label, name=""):
    if name != "":
        name = "/" + name
    prefix = "accuracy" + name
    pred = pred.view(-1)  # Binary accuracy
    label = label.view(-1)
    acc = label.float().eq((pred).float())
    return {prefix: as_float(acc.float().mean())}


@no_grad_func
def regression_accuracy(pred, label, name=""):
    if name != "":
        name = "/" + name
    prefix = "accuracy" + name
    pred = pred.view(-1)  # Binary accuracy
    label = label.view(-1)
    diff = pred - label
    return {
        prefix + "/l1": as_float(diff.abs().mean()),
        prefix + "/l2": as_float(0.5 * diff.pow(2).mean()),
    }


def _rms(p):
    return as_float((p ** 2).mean() ** 0.5)


@no_grad_func
def monitor_param_saturation(model):
    monitors = {}
    for name, p in model.named_parameters():
        p = F.sigmoid(p)
        sat = 1 - (p - (p > 0.5).float()).abs()
        monitors["sat/" + name] = sat
    return monitors


@no_grad_func
def monitor_param_rms(model):
    monitors = {}
    for name, p in model.named_parameters():
        monitors["param/rms/" + name] = _rms(p)
    return monitors


@no_grad_func
def monitor_param_gradrms(model):
    monitors = {}
    for name, p in model.named_parameters():
        if p.grad is not None:
            monitors["param/gradrms/" + name] = _rms(p.grad)
    return monitors


@no_grad_func
def monitor_param_gradrms_ratio(model):
    monitors = {}
    for name, p in model.named_parameters():
        if p.grad is not None:
            monitors["param/gradrmsratio/" + name] = _rms(p.grad) / max(_rms(p), 1e-8)
    return monitors



import torch

import jamtorch.prototype as jampt
from jammy.utils.gpu import gpu_by_weight
from jamtorch.logging import get_logger

__all__ = [
    "set_best_device",
]

logger = get_logger()
# from gpustat import GPUStatCollection

# import subprocess
# def get_pname(id):
#     p = subprocess.Popen(["ps -o cmd= {}".format(id)], stdout=subprocess.PIPE, shell=True)
#     return str(p.communicate()[0])

# def cuda_process():
#     gpu_stat = GPUStatCollection.new_query()


def set_best_device(mem_prior=1.0):
    gpu_id = gpu_by_weight(mem_prior)[0]
    logger.critical(f"select device: CUDA{gpu_id} ")
    torch.cuda.set_device(gpu_id)
    jampt.set_gpu_mode(True, gpu_id)
    return gpu_id



import pickle
from pathlib import Path

__all__ = ["AttrSaver"]


class AttrSaver:
    """
    class ABCSaver(AttrSaver):
        def __init__(self, cfg):
            super().__init__()
            self.save_path = cfg.path
            self._save_name = f"{cfg.name}_{cfg.seed:05d}"
            self.metric = {
                "metric_1": list(),
                "metric_2": dict()
            }
    """

    def __init__(self):
        self._save_name = "Foo"
        self._save_path = Path(".")

    @property
    def save_path(self):
        return self._save_path

    @save_path.setter
    def save_path(self, save_path):
        self._save_path = Path(save_path)
        self._save_path.mkdir(parents=True, exist_ok=True)

    def save(self):
        _file = f"{self._save_path}/{self._save_name}.pkl"
        with open(_file, "wb") as handle:
            pickle.dump(self, handle)

    @classmethod
    def load(cls, str_path):
        with open(str_path, "rb") as handle:
            learner = pickle.load(handle)
        return learner



import contextlib
import glob
import gzip
import json
import os
import os.path as osp
import pickle
import platform
import shutil
from zipfile import ZIP_DEFLATED, ZipFile

import numpy as np
import scipy.io as sio
import six
import yaml
from filelock import FileLock

from jammy.image import imread, imwrite
from jammy.logging import get_logger
from jammy.utils.enum import JamEnum
from jammy.utils.registry import CallbackRegistry, RegistryGroup

from .common import get_ext

logger = get_logger()

__all__ = [
    "as_file_descriptor",
    "fs_verbose",
    "set_fs_verbose",
    "open",
    "open_txt",
    "open_h5",
    "open_gz",
    "load",
    "load_txt",
    "load_h5",
    "load_pkl",
    "load_pklgz",
    "load_npy",
    "load_npz",
    "load_mat",
    "load_pth",
    "load_yaml",
    "load_json",
    "load_img",
    "dump",
    "dump_pkl",
    "dump_pklgz",
    "dump_npy",
    "dump_npz",
    "dump_mat",
    "dump_pth",
    "dump_json",
    "dump_img",
    "safe_dump",
    "compress",
    "compress_zip",
    "extract",
    "extract",
    "link",
    "mkdir",
    "lsdir",
    "remove",
    "locate_newest_file",
    "move",
    "copy",
    "replace",
    "io_function_registry",
    "latest_time",
]

sys_open = open


def as_file_descriptor(fd_or_fname, mode="r"):
    if isinstance(fd_or_fname, str):
        return sys_open(fd_or_fname, mode)  # pylint: disable=consider-using-with
    return fd_or_fname


def open_h5(file, mode, **kwargs):
    import h5py

    return h5py.File(file, mode, **kwargs)


def open_txt(file, mode, **kwargs):
    return sys_open(file, mode, **kwargs)  # pylint: disable=consider-using-with


def open_gz(file, mode):
    return gzip.open(file, mode)


def extract_zip(file, *args, **kwargs):
    with ZipFile(file, "r") as zip_ref:
        zip_ref.extractall(*args, **kwargs)


def load_pkl(file, **kwargs):
    with as_file_descriptor(file, "rb") as f:
        try:
            return pickle.load(f, **kwargs)
        except UnicodeDecodeError:
            if "encoding" in kwargs:
                raise
            return pickle.load(f, encoding="latin1", **kwargs)


# pylint: disable=unused-argument
def load_pklgz(file, **kwargs):
    with open_gz(file, "rb") as f:
        return load_pkl(f)


def load_h5(file, **kwargs):
    return open_h5(file, "r", **kwargs)


def load_txt(file, **kwargs):
    with sys_open(file, "r", **kwargs) as f:
        return f.readlines()


def load_npy(file, **kwargs):
    return np.load(file, **kwargs)


def load_npz(file, **kwargs):
    return np.load(file, **kwargs)


def load_mat(file, **kwargs):
    return sio.loadmat(file, **kwargs)


def load_pth(file, **kwargs):
    import torch

    return torch.load(file, **kwargs)


def load_json(file, **kwargs):
    with sys_open(file, "r") as fp:
        return json.load(fp)


def load_img(file, **kwargs):
    return imread(file)


def load_yaml(file, **kwargs):
    with sys_open(file, "r") as yamlfile:
        return yaml.load(yamlfile)


def dump_pkl(file, obj, **kwargs):
    with as_file_descriptor(file, "wb") as f:
        return pickle.dump(obj, f, **kwargs)


def dump_pklgz(file, obj, **kwargs):
    with open_gz(file, "wb") as f:
        return pickle.dump(obj, f)


def dump_yaml(file, obj, **kwargs):
    with sys_open(file, "w") as f:
        return yaml.dump(obj, f)


def dump_json(file, obj, **kwargs):
    with sys_open(file, "w") as f:
        return json.dump(obj, f, **kwargs)


def dump_img(file, obj, **kwargs):
    imwrite(file, obj)


def dump_npy(file, obj, **kwargs):
    return np.save(file, obj)


def dump_npz(file, obj, **kwargs):
    return np.savez(file, obj)


def dump_mat(file, obj, **kwargs):
    return sio.savemat(file, obj, **kwargs)


def dump_pth(file, obj, **kwargs):
    import torch

    return torch.save(obj, file)


def compress_zip(  # pylint: disable=inconsistent-return-statements
    file, file_list, verbose=True, **kwargs
):
    from jammy.cli import yes_or_no

    with ZipFile(file, "w", ZIP_DEFLATED) as cur_zip:
        for l_file in file_list:
            try:
                cur_zip.write(l_file)
            except FileNotFoundError:
                is_continue = yes_or_no(f"Missing {l_file}, continue?")
                if is_continue:
                    pass
                else:
                    return None


class _IOFunctionRegistryGroup(RegistryGroup):
    __base_class__ = CallbackRegistry

    def dispatch(self, registry_name, file, *args, **kwargs):
        entry = get_ext(file)
        callback = self.lookup(
            registry_name, entry, fallback=True, default=_default_io_fallback
        )
        return callback(file, *args, **kwargs)


def _default_io_fallback(file, *args, **kwargs):
    raise ValueError('Unknown file extension: "{}".'.format(file))


io_function_registry = _IOFunctionRegistryGroup()
io_function_registry.register("open", ".txt", open_txt)
io_function_registry.register("open", ".h5", open_h5)
io_function_registry.register("open", ".gz", open_gz)
io_function_registry.register("open", "__fallback__", sys_open)

io_function_registry.register("load", ".pkl", load_pkl)
io_function_registry.register("load", ".pklgz", load_pklgz)
io_function_registry.register("load", ".txt", load_txt)
io_function_registry.register("load", ".h5", load_h5)
io_function_registry.register("load", ".npy", load_npy)
io_function_registry.register("load", ".npz", load_npz)
io_function_registry.register("load", ".mat", load_mat)
io_function_registry.register("load", ".pth", load_pth)
io_function_registry.register("load", ".pt", load_pth)
io_function_registry.register("load", ".ckpt", load_pth)
io_function_registry.register("load", ".pt", load_pth)
io_function_registry.register("load", ".cfg", load_pkl)
io_function_registry.register("load", ".yaml", load_yaml)
io_function_registry.register("load", ".yml", load_yaml)
io_function_registry.register("load", ".json", load_json)
io_function_registry.register("load", ".jpg", load_img)
io_function_registry.register("load", ".png", load_img)
io_function_registry.register("load", ".jepg", load_img)

io_function_registry.register("dump", ".pkl", dump_pkl)
io_function_registry.register("dump", ".pklgz", dump_pklgz)
io_function_registry.register("dump", ".npy", dump_npy)
io_function_registry.register("dump", ".npz", dump_npz)
io_function_registry.register("dump", ".mat", dump_mat)
io_function_registry.register("dump", ".pth", dump_pth)
io_function_registry.register("dump", ".pt", dump_pth)
io_function_registry.register("dump", ".ckpt", dump_pth)
io_function_registry.register("dump", ".cfg", dump_pkl)
io_function_registry.register("dump", ".yaml", dump_yaml)
io_function_registry.register("dump", ".yml", dump_yaml)
io_function_registry.register("dump", ".json", dump_json)
io_function_registry.register("dump", ".jpg", dump_img)
io_function_registry.register("dump", ".png", dump_img)
io_function_registry.register("dump", ".jepg", dump_img)


io_function_registry.register("extract", ".zip", extract_zip)
io_function_registry.register("compress", ".zip", compress_zip)


_fs_verbose = False

# pylint: disable=global-statement
@contextlib.contextmanager
def fs_verbose(mode=True):
    global _fs_verbose

    _fs_verbose, mode = mode, _fs_verbose
    yield
    _fs_verbose = mode


def set_fs_verbose(mode=True):
    global _fs_verbose
    _fs_verbose = mode


def open(file, mode, **kwargs):  # pylint: disable=redefined-builtin
    if _fs_verbose and isinstance(file, six.string_types):
        logger.info('Opening file: "{}", mode={}.'.format(file, mode))
    return io_function_registry.dispatch("open", file, mode, **kwargs)


def load(file, **kwargs):
    if _fs_verbose and isinstance(file, six.string_types):
        logger.info('Loading data from file: "{}".'.format(file))
    return io_function_registry.dispatch("load", file, **kwargs)


def dump(file, obj, **kwargs):
    if _fs_verbose and isinstance(file, six.string_types):
        logger.info('Dumping data to file: "{}".'.format(file))
    return io_function_registry.dispatch("dump", file, obj, **kwargs)


def compress(file, obj, **kwargs):
    if _fs_verbose and isinstance(file, six.string_types):
        logger.info('compress data to file: "{}".'.format(file))
    return io_function_registry.dispatch("compress", file, obj, **kwargs)


def extract(file, **kwargs):
    if _fs_verbose and isinstance(file, six.string_types):
        logger.info('extract data to file: "{}".'.format(file))
    return io_function_registry.dispatch("extract", file, **kwargs)


def safe_dump(fname, data, use_lock=True, use_temp=True, lock_timeout=10):
    temp_fname = "temp." + fname
    lock_fname = "lock." + fname

    def safe_dump_inner():
        if use_temp:
            dump(temp_fname, data)
            os.replace(temp_fname, fname)
            return True
        else:
            return dump(temp_fname, data)

    if use_lock:
        with FileLock(lock_fname, lock_timeout) as flock:
            if flock.is_locked:
                return safe_dump_inner()
            else:
                logger.warning("Cannot lock the file: {}.".format(fname))
                return False
    else:
        return safe_dump_inner()


def link(path_origin, *paths, use_relative_path=True):
    for item in paths:
        if os.path.exists(item):
            os.remove(item)
        if use_relative_path:
            src_path = os.path.relpath(path_origin, start=os.path.dirname(item))
        else:
            src_path = path_origin
        try:
            os.symlink(src_path, item)
        except FileExistsError:
            os.unlink(item)
            os.symlink(src_path, item)


def mkdir(path):
    return os.makedirs(path, exist_ok=True)


class LSDirectoryReturnType(JamEnum):
    BASE = "base"
    NAME = "name"
    REL = "rel"
    FULL = "full"
    REAL = "real"


def lsdir(dirname, pattern=None, return_type="full"):
    assert "*" in dirname or "?" in dirname or osp.isdir(dirname)

    return_type = LSDirectoryReturnType.from_string(return_type)
    if pattern is not None:
        files = glob.glob(osp.join(dirname, pattern), recursive=True)
    elif "*" in dirname:
        files = glob.glob(dirname)
    else:
        files = os.listdir(dirname)

    if return_type is LSDirectoryReturnType.BASE:
        return [osp.basename(f) for f in files]
    elif return_type is LSDirectoryReturnType.NAME:
        return [osp.splitext(osp.basename(f))[0] for f in files]
    elif return_type is LSDirectoryReturnType.REL:
        assert (
            "*" not in dirname and "?" not in dirname
        ), "Cannot use * or ? for relative paths."
        return [osp.relpath(f, dirname) for f in files]
    elif return_type is LSDirectoryReturnType.FULL:
        return files
    elif return_type is LSDirectoryReturnType.REAL:
        return [osp.realpath(osp.join(dirname, f)) for f in files]
    else:
        raise ValueError("Unknown lsdir return type: {}.".format(return_type))


def remove(file):
    if osp.exists(file):
        if osp.isdir(file):
            shutil.rmtree(file, ignore_errors=True)
        if osp.isfile(file):
            os.remove(file)


def copy(src, dst):
    if osp.exists(src):
        if osp.isdir(src):
            _copy = shutil.copytree
        if osp.isfile(src):
            _copy = shutil.copyfile

        _copy(src, dst)


def move(src, dst):
    if osp.exists(src):
        os.rename(src, dst)


def replace(src, dst):
    if osp.exists(src):
        if osp.exists(dst):
            remove(dst)
        os.replace(src, dst)


def locate_newest_file(dirname, pattern):
    files = lsdir(dirname, pattern, return_type="full")
    if len(files) == 0:
        return None
    return max(files, key=osp.getmtime)


def latest_time(fname):
    import datetime

    if platform.system() == "Windows":
        ftime = os.path.getctime(fname)
    else:
        stat = os.stat(fname)
        try:
            ftime = stat.st_birthtime
        except AttributeError:
            # probably on Linux.
            ftime = stat.st_mtime
    return datetime.datetime.fromtimestamp(ftime)



import functools

import torch

__all__ = ["no_grad_func"]


def no_grad_func(func):
    @functools.wraps(func)
    def new_func(*args, **kwargs):
        with torch.no_grad():
            return func(*args, **kwargs)

    return new_func



# pylint: skip-file
# credit: https://github.com/Erotemic/xdev/blob/abdfa5271dbba9b73df8b9284e104ee2743912e1/xdev/profiler.py

from __future__ import absolute_import, division, print_function, unicode_literals
from six.moves import range, cStringIO
import six
import operator
import atexit
import sys
import re
import itertools as it
from collections import defaultdict

from .env import jam_getenv
__all__ = [
    'profile',
    'profile_now',
    'IS_PROFILING',
]

if '--profile' in sys.argv or jam_getenv("profile",type=bool):
    import line_profiler
    profile = line_profiler.LineProfiler()
    IS_PROFILING = True
else:
    def __dummy_profile__(func):
        """ dummy profiling func. does nothing """
        return func
    profile = __dummy_profile__
    IS_PROFILING = False


@atexit.register
def _dump_global_profile_report():
    # if we are profiling, then dump out info at the end of the program
    if IS_PROFILING:
        self = KernprofParser(profile)
        # print('----')
        # print('RAW')
        # print('----')
        # self.print_report()
        # print('----')
        # print('DUMPING')
        # print('----')
        self.dump_text()


def profile_now(func):
    """
    Eagerly report profile information after each call to `func`.

    Args:
        func (Callable): function to profile

    Example:
        >>> # xdoctest: +SKIP
        >>> from xdev.profiler import *  # NOQA
        >>> def func_to_profile():
        >>>     list(range(10))
        >>>     tuple(range(100))
        >>>     set(range(1000))
        >>> profile_now(func_to_profile)()  # xdoctest: +IGNORE_WANT
        Total time: 3.8e-05 s
        Function: func_to_profile at line 1
        Line #      Hits         Time  Per Hit   % Time  Line Contents
        ==============================================================
             1                                           def func_to_profile():
             2         1          4.0      4.0     10.5      list(range(10))
             3         1          3.0      3.0      7.9      tuple(range(100))
             4         1         31.0     31.0     81.6      set(range(1000))
    """
    import line_profiler
    profile = line_profiler.LineProfiler()
    new_func = profile(func)
    new_func.profile_info = KernprofParser(profile)
    new_func.print_report = new_func.profile_info.print_report

    def wraper(*args, **kwargs):
        try:
            return new_func(*args, **kwargs)
        except Exception:
            pass
        finally:
            new_func.print_report()
    wraper.new_func = new_func
    return wraper


class KernprofParser(object):

    def __init__(self, profile):
        self.profile = profile

    def raw_text(self):
        file_ = cStringIO()
        self.profile.print_stats(stream=file_, stripzeros=True)
        file_.seek(0)
        text =  file_.read()
        return text

    def print_report(self):
        print(self.raw_text())

    def get_text(self):
        text = self.raw_text()
        output_text, summary_text = self.clean_line_profile_text(text)
        return output_text, summary_text

    def dump_text(self):
        import ubelt as ub
        print("Dumping Profile Information")
        try:
            output_text, summary_text = self.get_text()
        except AttributeError:
            print('profile is not on')
        else:
            #profile.dump_stats('out.lprof')
            print(summary_text)
            suffix = ub.argval('--profname', default='')
            if suffix:
                suffix = '_' + suffix
            ub.writeto('profile_output{}.txt'.format(suffix), output_text + '\n' + summary_text)
            ub.writeto('profile_output{}.{}.txt'.format(suffix, ub.timestamp()),
                       output_text + '\n' + summary_text)

    def parse_rawprofile_blocks(self, text):
        """
        Split the file into blocks along delimters and and put delimeters back in
        the list
        """
        # The total time reported in the raw output is from pystone not kernprof
        # The pystone total time is actually the average time spent in the function
        delim = 'Total time: '
        delim2 = 'Pystone time: '
        #delim = 'File: '
        profile_block_list = re.split('^' + delim, text, flags=re.MULTILINE | re.DOTALL)
        for ix in range(1, len(profile_block_list)):
            profile_block_list[ix] = delim2 + profile_block_list[ix]
        return profile_block_list

    def clean_line_profile_text(self, text):
        """
        Sorts the output from line profile by execution time
        Removes entries which were not run
        """
        #
        profile_block_list = self.parse_rawprofile_blocks(text)
        #profile_block_list = fix_rawprofile_blocks(profile_block_list)
        #---
        # FIXME can be written much nicer
        prefix_list, timemap = self.parse_timemap_from_blocks(profile_block_list)
        # Sort the blocks by time
        sorted_lists = sorted(six.iteritems(timemap), key=operator.itemgetter(0))
        newlist = prefix_list[:]
        for key, val in sorted_lists:
            newlist.extend(val)
        # Rejoin output text
        output_text = '\n'.join(newlist)
        #---
        # Hack in a profile summary
        summary_text = self.get_summary(profile_block_list)
        output_text = output_text
        return output_text, summary_text

    def get_block_totaltime(self, block):

        def get_match_text(match):
            if match is not None:
                start, stop = match.start(), match.end()
                return match.string[start:stop]
            else:
                return None

        time_line = get_match_text(re.search('Pystone time: [0-9.]* s', block, flags=re.MULTILINE | re.DOTALL))
        if time_line is None:
            time_str = None
        else:
            time_str = get_match_text(re.search('[0-9.]+', time_line, flags=re.MULTILINE | re.DOTALL))
        if time_str is not None:
            return float(time_str)
        else:
            return None

    def get_block_id(self, block, readlines=None):

        def named_field(key, regex, vim=False):
            return r'(?P<%s>%s)' % (key, regex)

        non_unicode_whitespace = '[^ \t\n\r\f\v]'

        fpath_regex = named_field('fpath', non_unicode_whitespace + '+')
        funcname_regex = named_field('funcname', non_unicode_whitespace + '+')
        lineno_regex = named_field('lineno', '[0-9]+')

        fileline_regex = 'File: ' + fpath_regex + '$'
        funcline_regex = 'Function: ' + funcname_regex + ' at line ' + lineno_regex + '$'
        fileline_match = re.search(fileline_regex, block, flags=re.MULTILINE)
        funcline_match = re.search(funcline_regex, block, flags=re.MULTILINE)
        if fileline_match is not None and funcline_match is not None:
            fpath    = fileline_match.groupdict()['fpath']
            funcname = funcline_match.groupdict()['funcname']
            lineno   = funcline_match.groupdict()['lineno']
            # TODO: Determine if the function belongs to a class

            if readlines:
                # TODO: make robust
                classname = _find_parent_class(fpath, funcname, lineno, readlines)
                if classname:
                    funcname = classname + '.' + funcname
            block_id = funcname + ':' + fpath + ':' + lineno
        else:
            block_id = 'None:None:None'
        return block_id

    def parse_timemap_from_blocks(self, profile_block_list):
        """
        Build a map from times to line_profile blocks
        """
        prefix_list = []
        timemap = defaultdict(list)
        for ix in range(len(profile_block_list)):
            block = profile_block_list[ix]
            total_time = self.get_block_totaltime(block)
            # Blocks without time go at the front of sorted output
            if total_time is None:
                prefix_list.append(block)
            # Blocks that are not run are not appended to output
            elif total_time != 0:
                timemap[total_time].append(block)
        return prefix_list, timemap

    def get_summary(self, profile_block_list, maxlines=20):
        """
        References:
            https://github.com/rkern/line_profiler
        """
        import ubelt as ub
        time_list = [self.get_block_totaltime(block) for block in profile_block_list]
        time_list = [time if time is not None else -1 for time in time_list]

        @ub.memoize
        def readlines(fpath):
            return open(fpath, 'r').readlines()

        blockid_list = [self.get_block_id(block, readlines=readlines)
                        for block in profile_block_list]
        sortx = ub.argsort(time_list)
        sorted_time_list = list(ub.take(time_list, sortx))
        sorted_blockid_list = list(ub.take(blockid_list, sortx))

        aligned_blockid_list = _align_lines(sorted_blockid_list, ':')
        summary_lines = [('%6.2f seconds - ' % time) + line
                         for time, line in
                         zip(sorted_time_list, aligned_blockid_list)]

        summary_text = '\n'.join(summary_lines[-maxlines:])
        return summary_text

    def fix_rawprofile_blocks(self, profile_block_list):
        # TODO: finish function. should multiply times by
        # Timer unit to get true second profiling
        #profile_block_list_new = []
        for block in profile_block_list:
            block_lines = block.split('\n')
            sep = ['=' * 62]
            def split_block_at_sep(block_lines, sep):
                for pos, line in enumerate(block_lines):
                    if line.find(sep) == 0:
                        pos += 1
                        header_lines = block_lines[:pos]
                        body_lines = block_lines[pos:]
                        return header_lines, body_lines
                return block_lines, None
            header_lines, body_lines = split_block_at_sep(block_lines, sep)

    def clean_lprof_file(self, input_fname, output_fname=None):
        """ Reads a .lprof file and cleans it """
        # Read the raw .lprof text dump
        text = open(input_fname, 'r').read()
        # Sort and clean the text
        output_text = self.clean_line_profile_text(text)
        return output_text


def _find_parent_class(fpath, funcname, lineno, readlines=None):
    """
    Example:
        >>> from xdev import profiler
        >>> import ubelt as ub
        >>> funcname = 'clean_lprof_file'
        >>> func = getattr(profiler.KernprofParser, funcname)
        >>> lineno = func.__code__.co_firstlineno
        >>> fpath = profiler.__file__.replace('.pyc', '.py')
        >>> print('fpath = {!r}'.format(fpath))
        >>> #fpath = ub.truepath('~/code/xdev/xdev/profiler.py')
        >>> #lineno   = 264
        >>> readlines = lambda x: open(x, 'r').readlines()
        >>> classname = _find_parent_class(fpath, funcname, lineno, readlines)
        >>> print('classname = {!r}'.format(classname))
        >>> assert classname == 'KernprofParser', str(classname)
    """
    if readlines is None:
        def readlines(fpath):
            return open(fpath, 'r').readlines()

    try:
        line_list = readlines(fpath)
        row = int(lineno) - 1
        funcline = line_list[row]
        indent = len(funcline) - len(funcline.lstrip())
        if indent > 0:
            # get indentation
            # function is nested. fixme
            funcname = '<nested>:' + funcname
            # print('row = {!r}'.format(row))
            return _find_pyclass_above_row(line_list, row, indent)
    except Exception as ex:
        print('Got Error ex = {!r}'.format(ex))
        pass


def _find_pyclass_above_row(line_list, row, indent):
    """
    originally part of the vim plugin

    HACK: determine the class of the profiled funcs
    """
    import parse
    # Get text posision
    pattern = '^class [a-zA-Z_]'
    classline, classpos = _find_pattern_above_row(pattern, line_list, row,
                                                  indent, maxIter=None)
    result = parse.parse('class {name}({rest}', classline)
    classname = result.named['name']
    return classname


def _find_pattern_above_row(pattern, line_list, row, indent, maxIter=None):
    """
    searches a few lines above the curror until it **matches** a pattern
    """
    # Iterate until we match.
    # Janky way to find function / class name
    retval = None
    # print('row = {!r}'.format(row))
    # print('pattern = {!r}'.format(pattern))

    for ix in it.count(0):
        pos = row - ix
        if maxIter is not None and ix > maxIter:
            break
        if pos < 0:
            break
        searchline = line_list[pos]

        if indent is not None:
            if not searchline.strip():
                continue
            search_n_indent = len(searchline) - len(searchline.lstrip())
            if indent <= search_n_indent:
                continue
            # if indent < search_n_indent:
            #     continue

        if re.match(pattern, searchline) is not None:
            retval = searchline, pos
            break
    return retval


def _align_lines(line_list, character='=', replchar=None, pos=0):
    r"""
    Left justifies text on the left side of character

    TODO:
        clean up and move to ubelt?

    Args:
        line_list (list of strs):
        character (str):
        pos (int or list or None): does one alignment for all chars beyond this
            column position. If pos is None, then all chars are aligned.

    Returns:
        list: new_lines

    Example:
        >>> line_list = 'a = b\none = two\nthree = fish'.split('\n')
        >>> character = '='
        >>> new_lines = _align_lines(line_list, character)
        >>> result = ('\n'.join(new_lines))
        >>> print(result)
        a     = b
        one   = two
        three = fish

    Example:
        >>> line_list = 'foofish:\n    a = b\n    one    = two\n    three    = fish'.split('\n')
        >>> character = '='
        >>> new_lines = _align_lines(line_list, character)
        >>> result = ('\n'.join(new_lines))
        >>> print(result)
        foofish:
            a        = b
            one      = two
            three    = fish

    Example:
        >>> import ubelt as ub
        >>> character = ':'
        >>> text = ub.codeblock('''
            {'max': '1970/01/01 02:30:13',
             'mean': '1970/01/01 01:10:15',
             'min': '1970/01/01 00:01:41',
             'range': '2:28:32',
             'std': '1:13:57',}''').split('\n')
        >>> new_lines = _align_lines(text, ':', ' :')
        >>> result = '\n'.join(new_lines)
        >>> print(result)
        {'max'   : '1970/01/01 02:30:13',
         'mean'  : '1970/01/01 01:10:15',
         'min'   : '1970/01/01 00:01:41',
         'range' : '2:28:32',
         'std'   : '1:13:57',}

    Example:
        >>> line_list = 'foofish:\n a = b = c\n one = two = three\nthree=4= fish'.split('\n')
        >>> character = '='
        >>> # align the second occurence of a character
        >>> new_lines = _align_lines(line_list, character, pos=None)
        >>> print(('\n'.join(line_list)))
        >>> result = ('\n'.join(new_lines))
        >>> print(result)
        foofish:
         a   = b   = c
         one = two = three
        three=4    = fish
    """

    # FIXME: continue to fix ansi
    if pos is None:
        # Align all occurences
        num_pos = max([line.count(character) for line in line_list])
        pos = list(range(num_pos))

    # Allow multiple alignments
    if isinstance(pos, list):
        pos_list = pos
        # recursive calls
        new_lines = line_list
        for pos in pos_list:
            new_lines = _align_lines(new_lines, character=character,
                                     replchar=replchar, pos=pos)
        return new_lines

    # base case
    if replchar is None:
        replchar = character

    # the pos-th character to align
    lpos = pos
    rpos = lpos + 1

    tup_list = [line.split(character) for line in line_list]

    handle_ansi = True
    if handle_ansi:
        # Remove ansi from length calculation
        # References: http://stackoverflow.com/questions/14693701remove-ansi
        ansi_escape = re.compile(r'\x1b[^m]*m')

    # Find how much padding is needed
    maxlen = 0
    for tup in tup_list:
        if len(tup) >= rpos + 1:
            if handle_ansi:
                tup = [ansi_escape.sub('', x) for x in tup]
            left_lenlist = list(map(len, tup[0:rpos]))
            left_len = sum(left_lenlist) + lpos * len(replchar)
            maxlen = max(maxlen, left_len)

    # Pad each line to align the pos-th occurence of the chosen character
    new_lines = []
    for tup in tup_list:
        if len(tup) >= rpos + 1:
            lhs = character.join(tup[0:rpos])
            rhs = character.join(tup[rpos:])
            # pad the new line with requested justification
            newline = lhs.ljust(maxlen) + replchar + rhs
            new_lines.append(newline)
        else:
            new_lines.append(replchar.join(tup))
    return new_lines


import copy
import ctypes

import torch


# credit
# https://github.com/pyro-ppl/pyro/blob/beddc1f1a193d7f03bd588c9d6afe389bbd2d10c/pyro/distributions/util.py
class _DeepToMemo(dict):
    def __init__(self, to_args, to_kwargs):
        super().__init__()
        self._to_args = to_args
        self._to_kwargs = to_kwargs

    def get(self, key, default=None):
        result = super().get(key, default)
        if result is default:
            # Assume key is the id of another object, and look up that object.
            old = ctypes.cast(key, ctypes.py_object).value
            if isinstance(
                old, (torch.Tensor, torch.nn.Module)
            ):  # or maybe duck type here?
                self[key] = result = old.to(*self._to_args, **self._to_kwargs)
        return result


def deep_to(obj, *args, **kwargs):
    memo = _DeepToMemo(args, kwargs)
    return copy.deepcopy(obj, memo)



import os
import torch
import stackprinter
from filelock import FileLock, SoftFileLock
from .basic import get_rank, get_world_size, print0

# pylint: disable=broad-except

def try_filelock(
    fn,
    target_file,
    is_soft=False,
    timeout=20,
    device="cuda",
):
    lock_obj = SoftFileLock if is_soft else FileLock
    try:
        if get_rank() == 0:
            os.makedirs(os.path.dirname(target_file), exist_ok=True)
            lock = lock_obj(f"{target_file}._lock", timeout)
            lock.acquire()
            flag = lock.is_locked
        else:
            flag = False
        # create a one element tensor and broadcast it from rank 0
        flag = torch.tensor(flag, device=device)
        if get_world_size() > 1:
            torch.distributed.broadcast(flag, src=0)
        if flag.item():
            fn()
        else:
            print0(f"try_filelock Timeout!!! {target_file}")
    except TimeoutError:
        print0(f"TryFileLock timeout!!! {target_file}")
    except Exception as exc:
        stackprinter.show(exc, style="darkbg2")
    finally:
        if get_rank() == 0 and lock.is_locked:
            lock.release()

class TryFileLock:
    def __init__(self, target_file, is_soft=False, timeout=20, device="cuda"):
        self.target_file = target_file
        self.is_soft = is_soft
        self.timeout = timeout
        self.lock_obj = SoftFileLock if is_soft else FileLock
        self.lock = None
        self.device = device

    def __enter__(self):
        try:
            if get_rank() == 0:
                os.makedirs(os.path.dirname(self.target_file), exist_ok=True)
                self.lock = self.lock_obj(f"{self.target_file}._lock", self.timeout)
                self.lock.acquire()
                flag = self.lock.is_locked
            else:
                flag = False
            # create a one element tensor and broadcast it from rank 0
            flag = torch.tensor(flag, device=self.device)
            if get_world_size() > 1:
                torch.distributed.broadcast(flag, src=0)
            if flag.item():
                return self
            else:
                print0(f"TryFileLock can not acquire lock {self.target_file}")
                return None
        # handle timeout error
        except TimeoutError:
            # raise TimeoutError(f"TryFileLock timeout!!! {self.target_file}")
            print0(f"TryFileLock timeout!!! {self.target_file}")
            return None
        except Exception as exc:
            stackprinter.show(exc, style="darkbg2")
            raise RuntimeError(f"TryFileLock error!!!")
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        del exc_tb
        if exc_type is not None:
            print0(f"TryFileLock error!!! {exc_type} {exc_val}")
        if self.lock is not None and self.lock.is_locked:
            self.lock.release()


#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : cache.py
# Author : Jiayuan Mao
# Email  : maojiayuan@gmail.com
# Date   : 01/19/2018
#
# Qinsheng Zhang modified based on Jacinle.
# Distributed under terms of the MIT license.

import collections
import functools
import os.path as osp
import threading

from jammy.logging import get_logger

from .meta import synchronized

logger = get_logger()

__all__ = ["cached_property", "cached_result", "fs_cached_result"]


class cached_property:  # pylint: disable=too-few-public-methods, invalid-name
    def __init__(self, fget):
        self.fget = fget
        self.__module__ = fget.__module__
        self.__name__ = fget.__name__
        self.__doc__ = fget.__doc__
        self.__cache_key = "__result_cache_{}_{}".format(fget.__name__, id(fget))
        self.__mutex = collections.defaultdict(threading.Lock)

    def __get__(self, instance, owner):
        with self.__mutex[id(instance)]:
            if instance is None:
                return self.fget
            v = getattr(instance, self.__cache_key, None)
            if v is not None:
                return v
            v = self.fget(instance)
            assert v is not None
            setattr(instance, self.__cache_key, v)
            return v


def cached_result(func):
    def impl():
        nonlocal impl
        ret = func()
        impl = lambda: ret
        return ret

    @synchronized()
    @functools.wraps(func)
    def f():
        return impl()

    return f


def fs_cached_result(filename, force_update=False, verbose=False):
    import jammy.io as io

    def wrapper(func):
        @synchronized()
        @functools.wraps(func)
        def wrapped_func(*args, **kwargs):
            if not force_update and osp.exists(filename):
                if verbose:
                    logger.info('Using cached results from "{}".'.format(filename))
                cached_value = io.load(filename)
                if cached_value is not None:
                    return cached_value

            computed_value = func(*args, **kwargs)
            if verbose:
                logger.info('Writing result cache to "{}".'.format(filename))
            io.dump(filename, computed_value)
            return computed_value

        return wrapped_func

    return wrapper



from .ddp_trainer import DDPTrainer
from .fns import setup_ema

class EMATrainer(DDPTrainer):
    def __init__(self, cfg, loss_fn):
        super().__init__(cfg, loss_fn)
        setup_ema(cfg, self)

    def _impl_load_ckpt(self, state):
        if self.ema and "ema" in state:
            if state["ema"]:
                self.ema.load_state_dict(state["ema"])
        super()._impl_load_ckpt(state)

    def _impl_save_ckpt(self):
        rtn = {}
        if self.ema:
            rtn["ema"] = self.ema.state_dict()
        return {**super()._impl_save_ckpt(), **rtn}



import torch
import numpy as np

def gaussian_kde_log_eval(samples, query):
    N, D = samples.shape[0], samples.shape[-1]
    std = N ** (-1 / (D + 4))
    precision = (1 / (std ** 2)) * torch.eye(D)
    a = query - samples
    b = a @ precision
    c = -0.5 * torch.sum(a * b, dim=-1)
    d = -np.log(N) - (D / 2) * np.log(2 * np.pi) - D * np.log(std)
    c += d
    return torch.logsumexp(c, dim=-1)



import os
import smtplib
from email.mime.text import MIMEText
from email.utils import formataddr

from jammy.logging import get_logger

from .env import jam_getenv
from .meta import run_once

# pylint: disable=bare-except, no-member, no-self-use, broad-except,

logger = get_logger()

__all__ = ["jam_notifier", "get_notifier"]


class _Notifier:
    def __init__(self):
        self.enabled = True
        keys = ["sender", "receiver", "smtp_url", "smtp_port", "smtp_key"]
        for cur_key in keys:
            term = jam_getenv(f"ntf_{cur_key}", default=None)
            logger.info(f"get {term}")
            if term is None:
                self.enabled = False
                logger.info(f"miss {cur_key} in environment")
                break
            setattr(self, cur_key, term)
        try:
            user = os.getlogin()
        except:
            user = "qzhang419"
        self.address = f"{user}@{os.uname().nodename}"
        self.user = user

        self.server = smtplib.SMTP(self.smtp_url, self.smtp_port)
        self.server.login(self.sender, self.smtp_key)

    @run_once
    def warn(self):
        logger.critical("Notifier not setup")

    def notify(self, msg, subject=None):
        if not self.enabled:
            self.warn()
            return False
        logger.debug(f"send {msg}")
        try:
            msg = MIMEText(msg, "plain", "utf-8")
            msg["From"] = formataddr([f"NotifierEXP {self.address}", self.sender])
            msg["To"] = formataddr([f"{self.user}", self.receiver])
            if subject is None:
                subject = "Update on EXP"
            msg["Subject"] = subject

            self.server.sendmail(
                self.sender,
                [
                    self.receiver,
                ],
                msg.as_string(),
            )
            return True
        except Exception as e:
            logger.warning(str(e))
            return False

    def __exit__(self, *args, **kwargs):
        del args, kwargs
        self.server.quit()


jam_notifier = _Notifier()


def get_notifier():
    return jam_notifier



import numpy as np

__all__ = ["param_cnt"]


def param_cnt(model):
    model_parameters = filter(lambda p: p.requires_grad, model.parameters())
    return sum([np.prod(p.size()) for p in model_parameters])



import contextlib
import functools
import sys
import threading

from jammy.utils.env import jam_getenv

if jam_getenv("pdb", "ipdb").lower() == "ipdb":
    import ipdb as pdb
else:
    import pudb as pdb

__all__ = ["hook_exception_ipdb", "unhook_exception_ipdb", "exception_hook"]

# pylint: disable=invalid-name, redefined-builtin


def _custom_exception_hook(type, value, tb):
    if hasattr(sys, "ps1") or not sys.stderr.isatty():
        # we are in interactive mode or we don't have a tty-like
        # device, so we call the default hook
        sys.__excepthook__(type, value, tb)
    else:
        import traceback

        # we are NOT in interactive mode, print the exception...
        traceback.print_exception(type, value, tb)
        # ...then start the debugger in post-mortem mode.
        pdb.post_mortem(tb)


def hook_exception_ipdb():
    if not hasattr(_custom_exception_hook, "origin_hook"):
        _custom_exception_hook.origin_hook = sys.excepthook
        sys.excepthook = _custom_exception_hook


def unhook_exception_ipdb():
    assert hasattr(_custom_exception_hook, "origin_hook")
    sys.excepthook = _custom_exception_hook.origin_hook


@contextlib.contextmanager
def exception_hook(enable=True):
    if enable:
        hook_exception_ipdb()
        yield
        unhook_exception_ipdb()
    else:
        yield


def decorate_exception_hook(func):
    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        with exception_hook():
            return func(*args, **kwargs)

    return wrapped


def _TimeoutEnterIpdbThread(locals_, cv, timeout):
    del locals_
    with cv:
        if not cv.wait(timeout):

            pdb.set_trace()


@contextlib.contextmanager
def timeout_ipdb(locals_, timeout=3):
    cv = threading.Condition()
    thread = threading.Thread(
        target=_TimeoutEnterIpdbThread, args=(locals_, cv, timeout)
    )
    thread.start()
    yield
    with cv:
        cv.notify_all()



import os
import sys

from jammy.cli.keyboard import str2bool

from .cache import cached_result

__all__ = ["jam_getenv", "jam_is_verbose", "jam_is_debug", "jam_setenv"]


def jam_getenv(
    name, default=None, type=None, prefix=None
):  # pylint: disable=redefined-builtin
    if prefix is None:
        prefix = "JAM_"

    value = os.getenv((prefix + name).upper(), default)

    if value is None:
        return None

    if type is None:
        return value
    elif type == "bool":
        return str2bool(value)
    else:
        return type(value)


def jam_setenv(key, value, prefix=None):
    if prefix is None:
        prefix = "JAM_"
    os.environ[prefix + key.upper()] = str(value)


@cached_result
def jam_get_dashdebug_arg():
    # Return True if there is a '-debug' or '--debug' arg in the argv.
    for value in sys.argv:
        if value in ("-debug", "--debug"):
            return True
    return False


@cached_result
def jam_is_verbose(default="n", prefix=None):
    return jam_getenv("verbose", default, type="bool", prefix=prefix)


@cached_result
def jam_is_debug(default="n", prefix=None):
    return jam_get_dashdebug_arg() or jam_getenv(
        "debug", default, type="bool", prefix=prefix
    )



import functools
from contextlib import contextmanager
import torch.distributed as dist

__all__ = [
    "print0",
    "get_rank",
    "get_world_size",
    "is_master",
    "is_dist",
    "master_first",
    "master_only",
    "get_world_size",
    "barrier",
]

def print0(*args, **kwargs):
    if get_rank() == 0:
        print(*args, **kwargs)

def get_rank():
    return dist.get_rank() if dist.is_initialized() else 0

def get_world_size():
    return dist.get_world_size() if dist.is_initialized() else 1


def is_master():
    return not dist.is_initialized() or dist.get_rank() == 0


def is_dist():
    return dist.is_initialized()


def get_world_size():
    if is_dist():
        return dist.get_world_size()
    return 1


def barrier():
    if is_dist():
        dist.barrier()


@contextmanager
def master_first():
    if not is_master():
        barrier()
    yield
    if dist.is_initialized() and is_master():
        barrier()


def master_only(func):
    @functools.wraps(func)
    def new_func(*args, **kwargs):  # pylint: disable=inconsistent-return-statements
        if is_master():
            return func(*args, **kwargs)
        return

    return new_func


import threading
import inspect
import contextlib
import functools

from .meta import decorator_with_optional_args
from .naming import class_name_of_method

__all__ = [
    'defaults_manager', 'wrap_custom_as_default', 'gen_get_default', 'gen_set_default',
    'option_context',
    'ARGDEF', 'default_args'
]

class DefaultsManager(object):
    def __init__(self):
        self._is_local = dict()

        self._defaults_global = dict()
        self._defaults_local = threading.local()

    @decorator_with_optional_args(is_method=True)
    def wrap_custom_as_default(self, *, is_local=False):
        def wrapper(meth):
            identifier = class_name_of_method(meth)
            meth = contextlib.contextmanager(meth)
            self._is_local[identifier] = is_local
            defaults = self._get_defaults_registry(identifier)

            @contextlib.contextmanager
            @functools.wraps(meth)
            def wrapped_func(slf, *args, **kwargs):
                backup = defaults.get(identifier, None)
                defaults[identifier] = slf
                with meth(slf, *args, **kwargs):
                    yield
                defaults[identifier] = backup

            return wrapped_func
        return wrapper

    def gen_get_default(self, cls, default_getter=None):
        identifier = class_name_of_method(cls.as_default)

        def get_default(default=None):
            if default is None and default_getter is not None:
                default = default_getter()

            # NB(Jiayuan Mao): cannot use .get(identifier, default), because after calling as_default, the current
            #     default will be set to None.
            val = self._get_defaults_registry(identifier).get(identifier, None)
            if val is None:
                val = default
            return val
        return get_default

    def gen_set_default(self, cls):
        identifier = class_name_of_method(cls.as_default)

        def set_default(default):
            self._get_defaults_registry(identifier)[identifier] = default
        return set_default

    def set_default(self, cls, default):
        identifier = class_name_of_method(cls.as_default)
        self._get_defaults_registry(identifier)[identifier] = default

    def _get_defaults_registry(self, identifier):
        is_local = self._is_local.get(identifier, False)
        if is_local:
            if not hasattr(self._defaults_local, 'defaults'):
                self._defaults_local.defaults = dict()
            defaults = self._defaults_local.defaults
        else:
            defaults = self._defaults_global
        return defaults


defaults_manager = DefaultsManager()
wrap_custom_as_default = defaults_manager.wrap_custom_as_default
gen_get_default = defaults_manager.gen_get_default
gen_set_default = defaults_manager.gen_set_default

class _LocalObjectSimulator(object):
    __slots__ = ['ctx']


def option_context(name, is_local=True, **kwargs):
    class OptionContext(object):
        def __init__(self, **init_kwargs):
            for k, v in kwargs.items():
                setattr(self, k, v)
            if hasattr(self.__class__, 'current_context') and self.__class__.current_context.ctx is not None:
                c = self.__class__.get_default()
                for k in kwargs:
                    setattr(self, k, getattr(c, k))
            for k, v in init_kwargs.items():
                assert k in kwargs
                setattr(self, k, v)

        @classmethod
        def get_option(cls, name):
            getattr(cls.get_default(), name)

        @classmethod
        def set_default_option(cls, name, value):
            cls._create_default_context()
            setattr(cls.default_context.ctx, name, value)

        @classmethod
        def get_default(cls):
            cls._create_current_context()
            if cls.current_context.ctx is not None:
                return cls.current_context.ctx
            else:
                cls._create_default_context()
                return cls.default_context.ctx

        @contextlib.contextmanager
        def as_default(self):
            self.__class__._create_current_context()
            backup = self.__class__.current_context.ctx
            self.__class__.current_context.ctx = self
            yield
            self.__class__.current_context.ctx = backup

        @classmethod
        def _create_default_context(cls):
            if hasattr(cls, 'default_context'):
                return

            if is_local:
                cls.default_context = threading.local()
            else:
                cls.default_context = _LocalObjectSimulator()
            cls.default_context.ctx = cls(**kwargs)

        @classmethod
        def _create_current_context(cls):
            if hasattr(cls, 'current_context'):
                return

            if is_local:
                cls.current_context = threading.local()
            else:
                cls.current_context = _LocalObjectSimulator()
            cls.current_context.ctx = None

    OptionContext.__name__ = name

    return OptionContext



ARGDEF = object()


def default_args(func):
    def wrapper(func):
        sig = inspect.signature(func)

        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            bounded = sig.bind(*args, **kwargs)
            bounded.apply_defaults()

            for k, v in bounded.arguments.items():
                if v is ARGDEF:
                    if k in sig.parameters:
                        default_value = sig.parameters[k].default
                        bounded.arguments[k] = default_value

            return func(*bounded.args, **bounded.kwargs)

        return wrapped
    return wrapper(func)


#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : matching.py
# Author : Jiayuan Mao
# Email  : maojiayuan@gmail.com
# Date   : 03/02/2017
#
# Qinsheng modified based on Jacinle.
# Distributed under terms of the MIT license.


import collections
import fnmatch
import re

__all__ = ["NameMatcher", "IENameMatcher"]


class NameMatcher:
    def __init__(self, rules=None):
        if rules is None:
            self._rules = []
        elif isinstance(rules, dict):
            self._rules = list(rules.items())
        else:
            assert isinstance(rules, collections.Iterable)
            self._rules = list(rules)

        self._map = {}
        self._compiled_rules = []
        self._compiled = False

        self._matched = []
        self._unused = set()
        self._last_stat = None

    @property
    def rules(self):
        return self._rules

    def map(self):
        assert self._compiled
        return self._map

    def append_rule(self, rule):
        self._rules.append(tuple(rule))

    def insert_rule(self, index, rule):
        self._rules.insert(index, rule)

    def pop_rule(self, index=None):
        self._rules.pop(index)

    def begin(self, *, force_compile=False):
        if not self._compiled or force_compile:
            self.compile()
        self._matched = []
        self._unused = set(range(len(self._compiled_rules)))

    def end(self):
        return self._matched, {self._compiled_rules[i][0] for i in self._unused}

    def match(self, k):
        for i, (rule_str, re_rule, rule_value) in enumerate(self._compiled_rules):
            if re_rule.match(k):
                if i in self._unused:
                    self._unused.remove(i)
                self._matched.append((k, rule_str, rule_value))
                return rule_value
        return None

    def compile(self):
        self._map = dict()
        self._compiled_rules = []

        for rule_str, rule_value in self._rules:
            self._map[rule_str] = rule_value
            re_rule = fnmatch.translate(
                rule_str
            )  # shell-style pattern converted to regular expression
            re_rule = re.compile(re_rule, flags=re.IGNORECASE)
            self._compiled_rules.append((rule_str, re_rule, rule_value))
        self._compiled = True

    def __enter__(self):
        self.begin()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._last_stat = self.end()

    def get_last_stat(self):
        return self._last_stat


class IENameMatcher:
    def __init__(self, include=None, exclude=None):
        if include is None:
            self.include = None
        else:
            self.include = NameMatcher([(i, True) for i in include])

        if exclude is None:
            self.exclude = None
        else:
            self.exclude = NameMatcher([(e, True) for e in exclude])
        self._last_stat = None

    def begin(self):
        if self.include is not None:
            self.include.begin()
        if self.exclude is not None:
            self.exclude.begin()
        self._last_stat = (set(), set())

    def end(self):
        if self.include is not None:
            self.include.end()
        if self.exclude is not None:
            self.exclude.end()

        if len(self._last_stat[0]) < len(self._last_stat[1]):
            self._last_stat = ("included", self._last_stat[0])
        else:
            self._last_stat = ("excluded", self._last_stat[1])

    def match(self, k):
        if self.include is None:
            ret = True
        else:
            ret = bool(self.include.match(k))

        if self.exclude is not None:
            ret = ret and not bool(self.exclude.match(k))

        if ret:
            self._last_stat[0].add(k)
        else:
            self._last_stat[1].add(k)
        return ret

    def __enter__(self):
        self.begin()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.end()

    def get_last_stat(self):
        return self._last_stat



import functools

import numpy as np
import torch
from torch.utils import data
from tqdm.auto import tqdm

from jammy.random.rng import gen_rng
from jamtorch.utils.pytree import merge_tree

__all__ = ["get_batch", "get_subset", "num_to_groups"]


def get_batch_loader(data_loader):
    return next(iter(data_loader))


def get_batch(fdata, size=10):
    if isinstance(fdata, torch.utils.data.DataLoader):
        return get_batch_loader(fdata)
    if isinstance(fdata, torch.utils.data.Dataset):
        rng = gen_rng()
        subset = torch.utils.data.Subset(fdata, rng.choice(len(fdata), size=size))
        loader = torch.utils.data.DataLoader(subset, batch_size=size)
        return get_batch_loader(loader)

    raise RuntimeError(f"Type {type(fdata)} not supported")


def get_subset(dataset, target_size):
    if target_size is None:
        return dataset
    return data.Subset(dataset, np.random.choice(len(dataset), size=target_size))


def num_to_groups(num, divisor):
    groups = num // divisor
    remainder = num % divisor
    arr = [divisor] * groups
    if remainder > 0:
        arr.append(remainder)
    return arr


def batch_run(func, batch_size=100, is_tqdm=False):
    @functools.wraps(func)
    def new_func(total_batch, *args, **kwargs):
        arr = num_to_groups(total_batch, batch_size)
        rtn = []
        if is_tqdm:
            arr = tqdm(arr)
        for cur_batch_size in arr:
            rtn.append(func(cur_batch_size, *args, **kwargs))
        return merge_tree(rtn)

    return new_func



import six
import enum

__all__ = ["JamEnum"]


class JamEnum(enum.Enum):
    """A customized enumeration class, adding helper functions for string-based argument parsing."""

    @classmethod
    def from_string(cls, value):
        value = _canonize_enum_value(value, True)
        if isinstance(value, six.string_types) and hasattr(cls, value):
            return getattr(cls, value)
        value = _canonize_enum_value(value)
        return cls(value)

    @classmethod
    def type_name(cls):
        return cls.__name__

    @classmethod
    def choice_names(cls):
        return list(filter(lambda x: not x.startswith("_"), dir(cls)))

    @classmethod
    def choice_objs(cls):
        return [getattr(cls, name) for name in cls.choice_names()]

    @classmethod
    def choice_values(cls):
        return [getattr(cls, name).value for name in cls.choice_names()]

    @classmethod
    def is_valid(cls, value):
        value = _canonize_enum_value(value)
        return value in cls.choice_values()

    @classmethod
    def assert_valid(cls, value):
        assert cls.is_valid(value), 'Invalid {}: "{}". Supported choices: {}.'.format(
            cls.type_name(), value, ",".join(cls.choice_values())
        )

    def __jsonify__(self):
        return self.value


def _canonize_enum_value(value, cap=False):
    if isinstance(value, six.string_types):
        if cap:
            value = value.upper()
        else:
            value = value.lower()
    return value



import collections
import inspect
import os.path as osp
from io import StringIO

import numpy as np
import torch
import torch.nn as nn

import jammy.io as io
from jammy.utils.matching import IENameMatcher
from jammy.utils.printing import stprint
from jamtorch.logging import get_logger

__all__ = [
    "state_dict",
    "attr_dict",
    "save_ckpt",
    "load_ckpt",
    "aug_ckpt",
    "resume_cfg",
    "set_ckpt",
    "load_state_dict",
]


def state_dict(obj):
    if obj is None:
        return None
    if isinstance(obj, nn.Module):
        if hasattr(obj, "module"):
            obj = obj.module
    if hasattr(obj, "state_dict") and inspect.ismethod(obj.state_dict):
        return obj.state_dict()

    raise RuntimeError


def set_ckpt(src, des, atrs):
    for atr in atrs:
        setattr(des, atr, getattr(src, atr))


def attr_dict(obj, key_list):
    if isinstance(key_list, str):
        key_list = [key_list]
    if isinstance(
        key_list, (collections.abc.Set, collections.Sequence, collections.UserList)
    ):
        rtn = {}
        for key in key_list:
            rtn[key] = state_dict(getattr(obj, key))
        return rtn
    raise RuntimeError


def save_ckpt(state, is_best, filename="checkpoint", bestname="best"):
    filename = "{}.pth".format(filename)
    torch.save(state, filename)
    if is_best:
        io.link(filename, "{}.pth".format(bestname), use_relative_path=False)
        logger = get_logger()
        logger.info("Save best up to now: {}".format(filename))


def load_ckpt(gpu=None, filename="checkpoint"):
    logger = get_logger()
    for ckpt_file in [filename, f"{filename}.pth"]:
        if osp.isfile(ckpt_file):
            break
    if isinstance(gpu, int):
        gpu = f"cuda:{gpu}"
    if osp.isfile(ckpt_file):
        logger.critical("==> Loading from checkpoint '{}'".format(ckpt_file))
        checkpoint = torch.load(ckpt_file, map_location=gpu)
        if "env" in checkpoint and checkpoint["env"] is not None:
            env = checkpoint["env"]
            mem_buffer = StringIO()
            stprint(env, file=mem_buffer)
            logger.info("\n" + mem_buffer.getvalue())
        return checkpoint

    logger.critical("==> Checkpoint '{}' not found".format(filename))
    raise RuntimeError


def aug_ckpt(ckpt, aug_dict, is_save=False, ckpt_file=None):
    """augment the checkpoint,
    :returns: ckpt

    """
    if isinstance(ckpt, str):
        ckpt = torch.load(ckpt, torch.device("cpu"))

    assert isinstance(ckpt, dict)

    ckpt.update(aug_dict)

    if is_save:
        assert ckpt_file is not None
        torch.save(ckpt, ckpt_file)

    return ckpt


def resume_cfg(cfg):
    """resume cfg from ckpt"""
    if cfg.trainer is not None:
        n_cfg = cfg.trainer
    else:
        n_cfg = cfg

    if n_cfg.resume and n_cfg.ckpt is not None:
        ckpt_origin = n_cfg.ckpt
        for pth_file in [ckpt_origin, f"{ckpt_origin}.pth"]:
            if osp.isfile(pth_file):
                break
        ckpt = torch.load(pth_file, map_location=None)
        cfg = ckpt["cfg"] if "cfg" in ckpt else cfg
    return cfg


def load_state_dict(
    model, ckpt_state_dict, include=None, exclude=None
):  # pylint: disable= too-many-locals
    logger = get_logger()
    if hasattr(model, "module"):
        model = model.module

    matcher = IENameMatcher(include, exclude)
    with matcher:
        ckpt_state_dict = {k: v for k, v in ckpt_state_dict.items() if matcher.match(k)}
    stat = matcher.get_last_stat()  # examine unused rules
    if len(stat[1]) > 0:
        logger.critical(
            "Weights {}: {}.".format(stat[0], ", ".join(sorted(list(stat[1]))))
        )

    # Build the tensors.
    for k, v in ckpt_state_dict.items():
        if isinstance(v, np.ndarray):
            ckpt_state_dict[k] = torch.from_numpy(v)

    error_msg, warn_msg = [], []
    own_state = model.state_dict()
    for name, param in ckpt_state_dict.items():
        if name in own_state:
            if isinstance(param, nn.Parameter):
                # backwards compatibility for serialized parameters
                param = param.data
            try:
                own_state[name].copy_(param)
            except Exception:  # pylint: disable=broad-except
                error_msg.append(
                    "While copying the parameter named {}, "
                    "whose dimensions in the model are {} and "
                    "whose dimensions in the checkpoint are {}.".format(
                        name, own_state[name].size(), param.size()
                    )
                )

    missing = set(own_state.keys()) - set(ckpt_state_dict.keys())
    if len(missing) > 0:
        warn_msg.append('Missing keys in state_dict: "{}".'.format(missing))

    unexpected = set(ckpt_state_dict.keys()) - set(own_state.keys())
    if len(unexpected) > 0:
        warn_msg.append('Unexpected key "{}" in state_dict.'.format(unexpected))

    if len(error_msg) > 0:
        raise KeyError("\n".join(error_msg))

    if len(warn_msg) > 0:
        logger.critical("\n".join(warn_msg))



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : deprecated.py
# Author : Jiayuan Mao
# Email  : maojiayuan@gmail.com
# Date   : 04/21/2018
#
# This file is part of Jacinle.
# Distributed under terms of the MIT license.

import functools

from jammy.logging import get_logger

logger = get_logger()

__all__ = ["deprecated"]


def deprecated(func):
    @functools.wraps(func)
    def new_func(*args, **kwargs):
        if func not in deprecated.logged:
            deprecated.logged.add(func)
            logger.warning(func.__doc__)
        return func(*args, **kwargs)

    return new_func


deprecated.logged = set()



import sys

import numpy as np
import torch as th

from jammy.io.common import fsize_format
from jammy.utils.registry import CallbackRegistry

__all__ = ["sizeof"]

sys_sizeof = sys.getsizeof


def torch_sizeof(obj):
    return obj.element_size() * obj.nelement()


def np_sizeof(obj):
    return obj.nbytes


class _Registry(CallbackRegistry):
    def dispatch(self, name, *args, readable_string=True, **kwargs):
        size_in_byte = super().dispatch_direct(name, *args, **kwargs)
        if readable_string:
            return fsize_format(size_in_byte)
        return size_in_byte


_size_of_registry = _Registry()
_size_of_registry.set_fallback_callback(sys_sizeof)
_size_of_registry.register(int, sys_sizeof)
_size_of_registry.register(float, sys_sizeof)
_size_of_registry.register(complex, sys_sizeof)
_size_of_registry.register(np.ndarray, np_sizeof)
_size_of_registry.register(th.Tensor, torch_sizeof)


def sizeof(obj, readable_string=True):
    return _size_of_registry.dispatch(type(obj), obj, readable_string=readable_string)


if __name__ == "__main__":
    from jammy.utils.printing import stprint

    stprint(sizeof(1))
    stprint(sizeof(1.0))
    stprint(sizeof(1 + 1j))
    stprint(sizeof(np.zeros((2, 3))))
    stprint(sizeof(th.zeros((30000, 3, 256, 256))))



from .trainer import Trainer
from jamtorch.io import EMA
import torch.nn as nn

try:
    from apex import amp

    APEX_AVAILABLE = True
except:
    APEX_AVAILABLE = False

__all__ = ["CfgTrainer"]


class CfgTrainer(Trainer):
    def load_env(self, cfg):
        super().load_env(cfg)
        is_ema = cfg.get("ema") or False
        if is_ema:
            self.ema = EMA(
                cfg.get("ema_beta"),
                cfg.get("ema_num_warm"),
                cfg.get("ema_num_every"),
                self.model,
            )
            if cfg.get("ema_state") is not None:
                # if cfg from checkpoint
                self.ema.load_state_dict(cfg.get("ema_state"))
        else:
            self.ema = None
        self.load_fp16(cfg)

    def load_fp16(self, cfg):
        self.fp16 = bool(cfg.get("fp16") or False)
        assert not self.fp16 or self.fp16 and APEX_AVAILABLE, "INSTALL Apex"
        if self.fp16:
            if isinstance(self.model, nn.DataParallel):
                raise RuntimeError("DISABLE fp16 if using nn.DataParallel")

            if self.ema:
                (self.model, self.ema.model), self.optimizer = amp.initialize(
                    [self.model, self.ema.model], self.optimizer, opt_level="O1"
                )
            else:
                self.model, self.optimizer = amp.initialize(
                    self.model, self.optimizer, opt_level="O1"
                )

    def export_env(self):
        basic_state = super().export_env()
        new_state = {
            "ema": bool(self.ema),
            "ema_beta": self.ema.beta if self.ema else 0.9,
            "ema_num_warm": self.ema.num_warm if self.ema else 1,
            "ema_num_every": self.ema.num_every if self.ema else 1,
            "ema_state": self.ema.state_dict() if self.ema else None,
            "fp16": self.fp16,
        }
        return {**basic_state, **new_state}

    def load_cfg(self, cfg):
        self._cfg = cfg
        if cfg.resume:
            self.load_ckpt(cfg.ckpt, cfg.ckpt)
        else:
            self.load_env(cfg)

    def __call__(self, train_loader, val_loader):
        if self._cfg.resume:
            self.load_ckpt(self._cfg.ckpt)
            start_epoch = self.epoch_cnt + 1
            start_iter = self.iter_cnt + 1
            loss = self.best_loss
        else:
            start_epoch, start_iter, loss = 0, 0, float("inf")
        n_epochs = self._cfg.epochs
        return self.train(
            n_epochs, train_loader, val_loader, loss, start_iter, start_epoch
        )

    def train_step(self, feed_dict):
        rtn = super().train_step(feed_dict)
        if self.ema:
            self.ema.update_parameters(self.model)
        return rtn

    def loss_backward(self, loss):
        if self.fp16:
            with amp.scale_loss(loss, self.optimizer) as scaled_loss:
                scaled_loss.backward()
        else:
            loss.backward()
        return True



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : imgproc.py
# Author : Jiayuan Mao
# Email  : maojiayuan@gmail.com
# Date   : 01/19/2018
#
# This file is part of Jacinle.
# Distributed under terms of the MIT license.

import math
import functools

import numpy as np

from jammy.utils.argument import get_2dshape
from jammy.utils.enum import JamEnum

from . import backend

__all__ = [
    "resize",
    "resize_wh",
    "resize_scale",
    "resize_scale_wh",
    "resize_minmax",
    "crop",
    "center_crop",
    "leftup_crop",
    "dimshuffle",
    "clip",
    "clip_decorator",
    "grayscale",
    "brightness",
    "contrast",
    "saturation",
]


def _get_crop2d_rest(img, target_shape):
    source_shape = img.shape[:2]
    target_shape = get_2dshape(target_shape)
    rest_shape = source_shape[0] - target_shape[0], source_shape[1] - target_shape[1]
    assert rest_shape[0] >= 0 and rest_shape[1] >= 0
    return rest_shape


def _crop2d(img, start, size):
    return img[start[0] : start[0] + size[0], start[1] : start[1] + size[1]]


def resize(img, size, interpolation="LINEAR"):
    size = get_2dshape(size)
    return backend.resize(img, (size[1], size[0]), interpolation=interpolation)


def resize_wh(img, size_wh, interpolation="LINEAR"):
    size_wh = get_2dshape(size_wh)
    return backend.resize(img, size_wh, interpolation=interpolation)


def resize_scale(img, scale, interpolation="LINEAR"):
    scale = get_2dshape(scale, type=float)
    new_size = math.ceil(img.shape[0] * scale[0]), math.ceil(img.shape[1] * scale[1])
    return resize(img, new_size, interpolation=interpolation)


def resize_scale_wh(img, scale_wh, interpolation="LINEAR"):
    scale_wh = get_2dshape(scale_wh, type=float)
    return resize_scale(img, (scale_wh[1], scale_wh[0]), interpolation=interpolation)


def resize_minmax(img, min_dim, max_dim=None, interpolation="LINEAR"):
    if max_dim is None:
        max_dim = min_dim
    min_dim, max_dim = min(min_dim, max_dim), max(min_dim, max_dim)

    h, w = img.shape[:2]
    short, long = min(h, w), max(h, w)
    scale = min_dim / short
    scale = min(max_dim / long, scale)
    return resize_scale(img, scale, interpolation=interpolation)


def crop(image, l, t, w, h, extra_crop=None):
    if extra_crop is not None and extra_crop != 1:
        new_w, new_h = round(w * extra_crop), round(h * extra_crop)
        l -= (new_w - w) // 2
        t -= (new_h - h) // 2
        w, h = new_w, new_h

    im_h, im_w = image.shape[0:2]
    w, h = int(round(w)), int(round(h))
    l, t = int(math.floor(l)), int(math.floor(t))
    # range is expected to be image[t:t+h, l:l+w] now.

    ex_l, ex_t, ex_w, ex_h = l, t, w, h
    delta_l, delta_t = 0, 0
    if ex_l < 0:
        ex_l = 0
        delta_l = ex_l - l
        ex_w -= delta_l
    if ex_t < 0:
        ex_t = 0
        delta_t = ex_t - t
        ex_h -= delta_t
    if ex_l + ex_w > im_w:
        ex_w = im_w - ex_l
    if ex_t + ex_h > im_h:
        ex_h = im_h - ex_t

    result = np.zeros(shape=(h, w) + image.shape[2:], dtype=image.dtype)
    result[delta_t : delta_t + ex_h, delta_l : delta_l + ex_w] = image[
        ex_t : ex_t + ex_h, ex_l : ex_l + ex_w
    ]
    return result


def center_crop(img, target_shape):
    """ center crop """
    target_shape = get_2dshape(target_shape)
    rest = _get_crop2d_rest(img, target_shape)
    start = rest[0] // 2, rest[1] // 2

    return _crop2d(img, start, target_shape)


def leftup_crop(img, target_shape):
    """ left-up crop """
    start = 0, 0
    target_shape = get_2dshape(target_shape)

    return _crop2d(img, start, target_shape)


class ShuffleType(JamEnum):
    CHANNEL_FIRST = "channel_first"
    CHANNEL_LAST = "channel_last"


def dimshuffle(img, shuffle_type):
    shuffle_type = ShuffleType.from_string(shuffle_type)
    assert len(img.shape) in (2, 3, 4), "Image should be of dims 2, 3 or 4"

    if len(img.shape) == 2:
        return img
    elif len(img.shape) == 3:
        if shuffle_type == ShuffleType.CHANNEL_FIRST:
            return np.transpose(img, (2, 0, 1))
        else:
            return np.transpose(img, (1, 2, 0))
    else:  # len(img.shape) == 4:
        if shuffle_type == ShuffleType.CHANNEL_FIRST:
            return np.transpose(img, (0, 3, 1, 2))
        else:
            return np.transpose(img, (0, 2, 3, 1))


def clip(img):
    return np.minimum(255, np.maximum(0, img))


def clip_decorator(func):
    @functools.wraps(func)
    def new_func(*args, **kwargs):
        img = func(*args, **kwargs)
        return clip(img)

    return new_func


def grayscale(img):
    assert len(img.shape) == 3 and img.shape[2] == 3
    w = np.array([0.114, 0.587, 0.299]).reshape(1, 1, 3)
    img = (img * w).sum(axis=2, keepdims=True)
    return img


@clip_decorator
def brightness(img, alpha):
    return img * alpha


@clip_decorator
def contrast(img, alpha):
    gs = grayscale(img)
    gs[:] = gs.mean()
    img = img * alpha + gs * (1 - alpha)
    return img


@clip_decorator
def saturation(img, alpha):
    gs = grayscale(img)
    img = img * alpha + gs * (1 - alpha)
    return img



from torch import nn

__all__ = [
    "Discriminator",
]

# pylint: disable=invalid-name, too-many-arguments


def _downsample(x):
    # Downsample (Mean Avg Pooling with 2x2 kernel)
    return nn.AvgPool2d(kernel_size=2)(x)


class OptimizedDisBlock(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        ksize=3,
        pad=1,
        d_spectral_norm=False,
        activation=nn.ReLU(),
    ):
        super().__init__()
        self.activation = activation

        self.c1 = nn.Conv2d(in_channels, out_channels, kernel_size=ksize, padding=pad)
        self.c2 = nn.Conv2d(out_channels, out_channels, kernel_size=ksize, padding=pad)
        self.c_sc = nn.Conv2d(in_channels, out_channels, kernel_size=1, padding=0)
        if d_spectral_norm:
            self.c1 = nn.utils.spectral_norm(self.c1)
            self.c2 = nn.utils.spectral_norm(self.c2)
            self.c_sc = nn.utils.spectral_norm(self.c_sc)

    def residual(self, x):
        h = x
        h = self.c1(h)
        h = self.activation(h)
        h = self.c2(h)
        h = _downsample(h)
        return h

    def shortcut(self, x):
        return self.c_sc(_downsample(x))

    def forward(self, x):
        return self.residual(x) + self.shortcut(x)


class DisBlock(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        hidden_channels=None,
        ksize=3,
        pad=1,
        d_spectral_norm=False,
        activation=nn.ReLU(),
        downsample=False,
    ):
        super().__init__()
        self.activation = activation
        self.downsample = downsample
        self.learnable_sc = (in_channels != out_channels) or downsample
        hidden_channels = in_channels if hidden_channels is None else hidden_channels
        self.c1 = nn.Conv2d(
            in_channels, hidden_channels, kernel_size=ksize, padding=pad
        )
        self.c2 = nn.Conv2d(
            hidden_channels, out_channels, kernel_size=ksize, padding=pad
        )
        if d_spectral_norm:
            self.c1 = nn.utils.spectral_norm(self.c1)
            self.c2 = nn.utils.spectral_norm(self.c2)

        if self.learnable_sc:
            self.c_sc = nn.Conv2d(in_channels, out_channels, kernel_size=1, padding=0)
            if d_spectral_norm:
                self.c_sc = nn.utils.spectral_norm(self.c_sc)

    def residual(self, x):
        h = x
        h = self.activation(h)
        h = self.c1(h)
        h = self.activation(h)
        h = self.c2(h)
        if self.downsample:
            h = _downsample(h)
        return h

    def shortcut(self, x):
        if self.learnable_sc:
            x = self.c_sc(x)
            if self.downsample:
                return _downsample(x)
            else:
                return x
        else:
            return x

    def forward(self, x):
        return self.residual(x) + self.shortcut(x)


class Discriminator(nn.Module):
    def __init__(
        self, df_dim, in_channel=3, d_spectral_norm=False, activation=nn.PReLU()
    ):
        super().__init__()
        self.ch = df_dim
        self.activation = activation
        self.block1 = OptimizedDisBlock(
            in_channel, self.ch, d_spectral_norm=d_spectral_norm
        )
        self.block2 = DisBlock(
            self.ch,
            self.ch,
            d_spectral_norm=d_spectral_norm,
            activation=activation,
            downsample=True,
        )
        self.block3 = DisBlock(
            self.ch,
            self.ch,
            d_spectral_norm=d_spectral_norm,
            activation=activation,
            downsample=False,
        )
        self.block4 = DisBlock(
            self.ch,
            self.ch,
            d_spectral_norm=d_spectral_norm,
            activation=activation,
            downsample=False,
        )
        self.l5 = nn.Linear(self.ch, 1, bias=False)
        if d_spectral_norm:
            self.l5 = nn.utils.spectral_norm(self.l5)

    def forward(self, x):
        h = x
        h = self.block1(h)
        h = self.block2(h)
        h = self.block3(h)
        h = self.block4(h)
        h = self.activation(h)
        # Global average pooling
        h = h.sum(2).sum(2)
        output = self.l5(h)

        return output



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : kv.py
# Author : Jiayuan Mao
# Email  : maojiayuan@gmail.com
# Date   : 01/19/2018
#
# This file is part of Jacinle.
# Distributed under terms of the MIT license.

from jammy.utils.context import EmptyContext


class KVStoreBase:
    def __init__(self, readonly=False):
        self.__readonly = readonly

    @property
    def readonly(self):
        return self.__readonly

    def has(self, key, **kwargs):
        return self._has(key, **kwargs)

    def get(self, key, default=None, **kwargs):
        return self._get(key, default=default, **kwargs)

    def put(self, key, value, replace=True, **kwargs):
        assert not self.readonly, "KVStore is readonly: {}.".format(self)
        return self._put(key, value, replace=replace, **kwargs)

    def update(self, key, value, **kwargs):
        kwargs["replace"] = True
        self.put(key, value, **kwargs)

    def erase(self, key, **kwargs):
        assert not self.readonly, "KVStore is readonly: {}.".format(self)
        return self._erase(key, **kwargs)

    def __contains__(self, key):
        return self.has(key)

    def __getitem__(self, key):
        return self.get(key)

    def __setitem__(self, key, value):
        return self.set(key, value)  # pylint: disable=no-member

    def __delitem__(self, key):
        return self.erase(key)

    def transaction(self, *args, **kwargs):
        return self._transaction(*args, **kwargs)

    def keys(self):
        return self._keys()

    def _has(self, key):
        raise NotImplementedError(
            "KVStore {} does not support has.".format(self.__class__.__name__)
        )

    def _get(self, key, default):
        raise NotImplementedError(
            "KVStore {} does not support get.".format(self.__class__.__name__)
        )

    def _put(self, key, value, replace):
        raise NotImplementedError(
            "KVStore {} does not support put.".format(self.__class__.__name__)
        )

    def _erase(self, key):
        raise NotImplementedError(
            "KVStore {} does not support erase.".format(self.__class__.__name__)
        )

    def _transaction(
        self, *args, **kwargs
    ):  # pylint: disable=unused-argument,no-self-use
        return EmptyContext()

    def _keys(self):
        raise NotImplementedError(
            "KVStore {} does not support keys access.".format(self.__class__.__name__)
        )



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : memcached.py
# Author : Jiayuan Mao
# Email  : maojiayuan@gmail.com
# Date   : 01/19/2018
#
# Qinsheng Zhang modified based on Jacinle.
# Distributed under terms of the MIT license.
# pylint: skip-file
import pickle

try:
    import memcache
except ImportError:
    from jammy.logging import get_logger

    logger = get_logger()
    logger.warning(
        "Cannot import memcache. MemcachedKVStore is unavailable. Install python-memcached"
    )
    memcache = None

from .kv import KVStoreBase

_loads = pickle.loads
_dumps = pickle.dumps
_encode = lambda s: s.encode("utf8")


class MemcachedKVStore(KVStoreBase):
    def __init__(self, addr, port, readonly=False):
        super().__init__(readonly)
        self.available = memcache is not None

        self.addr = addr
        self.port = port
        self.full_addr = "{}:{}".format(addr, port)

        self._connection = None

    @property
    def connection(self):
        if self._connection is None:
            self._connection = memcache.Client([self.full_addr], debug=0)
        return self._connection

    def _has(self, key):
        key = _encode(key)
        try:
            data = self.connection.get(key)
            if data is not None:
                return True
            return False
        except IOError:
            return False

    def _get(self, key, default=None, refresh=False, refresh_timeout=0):
        key = _encode(key)
        try:
            data = self.connection.get(key)
            if data is not None:
                if refresh:
                    self.connection.replace(key, data, refresh_timeout, 1)
                return _loads(data)
            else:
                return default
        except IOError:
            return default

    def _put(self, key, value, replace=False, timeout=0):
        key = _encode(key)
        # TODO(Jiayuan Mao @ 10/23): implement replace.
        assert replace, "Not implemented."
        self.connection.set(key, _dumps(value), timeout, 1)

    def _erase(self, key):
        key = _encode(key)
        self.connection.delete(key)



import os
import random as sys_random

import numpy as np
import numpy.random as npr

from jammy.utils.cache import cached_result
from jammy.utils.defaults import defaults_manager
from jammy.utils.env import jam_getenv
from jammy.utils.registry import Registry

__all__ = [
    "JamRandomState",
    "get_default_rng",
    "gen_seed",
    "gen_rng",
    "reset_global_seed",
    "jam_rng_seed",
]


class JamRandomState(npr.RandomState):
    def choice_list(self, list_, size=1, replace=False, p=None):
        """Efficiently draw an element from an list,
        if the rng is given, use it instead of the system one."""
        if size == 1:
            if type(list_) in (list, tuple):
                return list_[self.choice(len(list_), p=p)]
            return self.choice(list_, p=p)
        else:
            if type(list_) in (list, tuple):
                inds = self.choice(len(list_), size=size, replace=replace, p=p)
                return [list_[i] for i in inds]
            return self.choice(list_, size=size, replace=replace, p=p)

    def shuffle_list(self, list_):
        if isinstance(list_, list):
            sys_random.shuffle(list_, random=self.random_sample)
        else:
            self.shuffle(list_)

    def shuffle_multi(self, *arrs):
        length = len(arrs[0])
        for cur_a in arrs:
            assert (
                len(cur_a) == length
            ), "non-compatible length when shuffling multiple arrays"

        inds = np.arange(length)
        self.shuffle(inds)
        return tuple(map(lambda x: x[inds], arrs))

    @defaults_manager.wrap_custom_as_default(is_local=True)
    def as_default(self):
        yield self


@cached_result
def jam_rng_seed():
    return jam_getenv("RANDOM_SEED", 8, int)


_rng = JamRandomState()


get_default_rng = defaults_manager.gen_get_default(
    JamRandomState, default_getter=lambda: _rng
)


class _RngRegistry(Registry):
    def register(self, entry, value):
        seed = jam_rng_seed()
        if seed is not None:
            value()(seed)
        return super().register(entry, value)


def gen_seed():
    return get_default_rng().randint(4294967296)


def gen_rng(seed=None):
    if seed is None:
        seed = jam_rng_seed()
    return JamRandomState(seed)


global_rng_registry = _RngRegistry()
global_rng_registry.register("jammy", lambda: _rng.seed)
global_rng_registry.register("numpy", lambda: npr.seed)
global_rng_registry.register("sys", lambda: sys_random.seed)


def reset_global_seed(seed=None, verbose=True):
    if seed is None:
        seed = gen_seed()
    seed = int(seed)
    for k, seed_getter in global_rng_registry.items():
        if verbose:
            from jammy.logging import get_logger

            logger = get_logger()
            logger.critical(
                "Reset random seed for: {} (pid={}, seed={}).".format(
                    k, os.getpid(), seed
                )
            )
        seed_getter()(seed)



import numpy as np
import matplotlib.pyplot as plt

__all__ = [
    "ndimgs_in_row",
]

def ndimgs_in_row(list_x, n, dpi=128, img_size=400):
    """
    args:
    list_x: list of elements that imshow can display
    n: number of the element in a row
    """
    length = len(list_x)
    idxes = np.linspace(0, length - 1, n, dtype=int)
    with plt.style.context("img"):
        fig, axs = plt.subplots(
            1, n, figsize=(n * img_size / dpi, img_size / dpi), dpi=dpi
        )
        for j, idx in enumerate(idxes):
            axs[j].imshow(list_x[idx])
            axs[j].set_title(idx)
    return fig


from jamtorch.utils.meta import is_master
from jamtorch.io import hyd_ema

__all__ = ["setup_ema"]

def setup_ema(cfg, trainer):
    trainer.ema = None
    # FIXME: check ema_master key word design
    if cfg.ema_master: # only one ema on the master
        if is_master():
            trainer.ema = hyd_ema(cfg)
    else:
        trainer.ema = hyd_ema(cfg)

    if trainer.ema:
        def update_ema(_trainer):
            trainer.ema.update_parameters(trainer.model)
        
        trainer.register_event("step:end", update_ema, False)





import os
import sys
import select
import os.path as osp

__all__ = ["str2bool", "yn2bool", "yes_or_no", "maybe_mkdir", "timeout_input"]


def str2bool(s):
    if s.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif s.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise ValueError('str2bool is undefined for: "{}".'.format(s))


def yn2bool(s):
    if s.lower() in ("yes", "y"):
        return True
    elif s.lower() in ("no", "n"):
        return False
    else:
        raise ValueError('yn2bool is undefined for: "{}".'.format(s))


def yes_or_no(question, default="yes"):
    """Ask a yes/no question via raw_input() and return their answer.
    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
    It must be "yes" (the default), "no" or None (meaning that
    an answer is required from the user).
    The "answer" return value is True for "yes" or False for "no".
    """

    valid = {
        "yes": True,
        "y": True,
        "ye": True,
        "no": False,
        "n": False,
        "default": None,
        "def": None,
        "d": None,
    }

    quiet = os.getenv("JAM_QUIET", "")
    if quiet != "":
        quiet = quiet.lower()
        assert quiet in valid, "Invalid JAM_QUIET environ: {}.".format(quiet)
        choice = valid[quiet]
        sys.stdout.write(
            "Jam Quiet run:\n\tQuestion: {}\n\tChoice: {}\n".format(
                question, "Default" if choice is None else "Yes" if choice else "No"
            )
        )
        return choice if choice is not None else default

    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("Invalid default answer: '%s'." % default)

    while True:
        sys.stdout.write(question + prompt)
        choice = input().lower()
        if default is not None and choice == "":
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' " "(or 'y' or 'n').\n")


def maybe_mkdir(dirname):
    if osp.isdir(dirname):
        return
    if osp.isfile(dirname):
        return

    import jammy.io as io

    if yes_or_no('Creating directory "{}"?'.format(dirname)):
        io.mkdir(dirname)


def timeout_input(prompt, timeout=10, default="", strict=False):
    print(prompt, end=": ", flush=True)
    inputs, outputs, errors = select.select([sys.stdin], [], [], timeout)
    print()
    if not inputs and strict:
        raise RuntimeError
    return sys.stdin.readline().strip()



from jammy.utils import git
from jammy.utils.env import jam_getenv
from jammy.utils.init import init_main

init_main()

del init_main

__hash__ = git.git_hash(__file__)

if jam_getenv("IMPORT_ALL", "true", "bool"):
    from jammy import io
    from jammy.utils.cnt import BufferCnt, CBCnt, bufcnt
    from jammy.utils.hyd import hyd_instantiate, hydpath
    from jammy.utils.hyd import instantiate as jam_instantiate
    from jammy.utils.hyd import link_hyd_run, update_cfg
    from jammy.utils.imp import load_class, load_module
    from jammy.utils.meta import (
        Singleton,
        assert_instance,
        assert_none,
        assert_notnone,
        cond_with,
        cond_with_group,
        decorator_with_optional_args,
        dict_deep_keys,
        dict_deep_kv,
        dict_deep_update,
        filter_exec,
        first_n,
        gofor,
        make_dummy_func,
        map_exec,
        map_exec_method,
        merge_iterable,
        method2func,
        notnone_property,
        run_once,
        stmap,
        synchronized,
        try_run,
    )
    from jammy.utils.printing import stprint


def get_jam_repo_git():
    """get git info running jammy

    :return: (jam_sha, jam_diff)
    """
    return git.log_repo(__file__)



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : imgio.py
# Author : Jiayuan Mao
# Email  : maojiayuan@gmail.com
# Date   : 01/19/2018
#
# This file is part of Jacinle.
# Distributed under terms of the MIT license.


import os as os
import os.path as osp
import numpy as np

from . import backend
from .imgproc import dimshuffle


__all__ = [
    "imread",
    "imwrite",
    "imshow",
    "plt2pil",
    "plt2nd",
    "nd2pil",
    "pil2nd",
    "savefig",
]


def imread(path, *, shuffle=False):
    if not osp.exists(path):
        return None
    i = backend.imread(path)
    if i is None:
        return None
    if shuffle:
        return dimshuffle(i, "channel_first")
    return i


def imwrite(path, img, *, shuffle=False):
    if shuffle:
        img = dimshuffle(img, "channel_last")
    backend.imwrite(path, img)


def imshow(title, img, *, shuffle=False):
    if shuffle:
        img = dimshuffle(img, "channel_last")
    backend.imshow(title, img)


def plt2pil(fig):
    """Convert a Matplotlib figure to a PIL Image and return it"""
    import io
    from PIL import Image

    buf = io.BytesIO()
    fig.savefig(buf)
    buf.seek(0)
    img = Image.open(buf)
    return img

def plt2nd(fig):
    return np.array(plt2pil(fig))

def savefig(fig, fig_name):
    fig_path = fig_name.split("/")
    if len(fig_path) > 1:
        save_path = "/".join(fig_path[:-1])
        if not osp.isdir(save_path):
            os.makedirs(save_path, exist_ok=True)
    fig.savefig(fig_name)


nd2pil=backend.pil_nd2img
pil2nd=backend.pil_img2nd